"""
CPU EMULATOR HANDLED INSTRUCTIONS

Add any instructions that need to be handled below.  The function should be declared as such

# Using the same function for multiple instructions:
@opcode("add")
@opcode("adc")
def _add(cpu_context, instruction):
    print "IN ADD"

# Using a single function for an opcode
@opcode
def MOV(cpu_context, instruction):
    print "IN MOV"

WARNING:
    Do NOT rely on the flags registers being correct.  There are places were flags are NOT being updated when they
    should, and the very fact that CALL instructions are skipped could cause flags to be incorrect.
"""

import logging

import dragodis
from dragodis import OperandType

from .. import utils
from ..cpu_context import ProcessorContext
from ..instruction import Instruction
from ..registry import registrar

# Dictionary containing opcode names -> function
OPCODES = {}
opcode = registrar(OPCODES, name="opcode")


logger = logging.getLogger(__name__)


@opcode
def AAA(cpu_context: ProcessorContext, instruction: Instruction):
    """ ASCII Adjust AX After Addition """
    if cpu_context.bitness == 64:
        logger.debug("Opcode not valid for 64-bit")
        return

    orig_ax = cpu_context.registers.ax

    if (cpu_context.registers.al & 0xF) > 9 or cpu_context.registers.af == 1:
        cpu_context.registers.ax += 0x106
        cpu_context.registers.af = 1
        cpu_context.registers.cf = 1
    else:
        cpu_context.registers.af = 0
        cpu_context.registers.cf = 0
    cpu_context.registers.al &= 0xF

    logger.debug("Adjusted AX 0x%X -> 0x%X", orig_ax, cpu_context.registers.ax)


@opcode
def AAD(cpu_context: ProcessorContext, instruction: Instruction):
    """ ASCII Adjust AX Before Division """
    if cpu_context.bitness == 64:
        logger.debug("Opcode not valid for 64-bit")
        return

    orig_ax = cpu_context.registers.ax

    operands = instruction.operands
    base = operands[0].value if operands else 10
    al = cpu_context.registers.al
    ah = cpu_context.registers.ah
    cpu_context.registers.al = (al + (ah * base)) & 0xFF
    cpu_context.registers.ah = 0

    logger.debug("Adjusted AX 0x%X -> 0x%X", orig_ax, cpu_context.registers.ax)


@opcode
def AAM(cpu_context: ProcessorContext, instruction: Instruction):
    """ ASCII Adjust AX After Multiply """
    if cpu_context.bitness == 64:
        logger.debug("Opcode not valid for 64-bit")
        return

    orig_ax = cpu_context.registers.ax

    operands = instruction.operands
    base = operands[0].value if operands else 10
    al = cpu_context.registers.al
    cpu_context.registers.ah = al // base
    cpu_context.registers.al = al % base

    logger.debug("Adjusted AX 0x%X -> 0x%X", orig_ax, cpu_context.registers.ax)


@opcode
def AAS(cpu_context: ProcessorContext, instruction: Instruction):
    """ ASCII Adjust AX After Subtraction """
    if cpu_context.bitness == 64:
        logger.debug("Opcode not valid for 64-bit")
        return

    orig_ax = cpu_context.registers.ax

    if (cpu_context.registers.al & 0xF) > 9 or cpu_context.registers.af == 1:
        cpu_context.registers.ax -= 6
        cpu_context.registers.ah -= 1
        cpu_context.registers.af = 1
        cpu_context.registers.cf = 1
    else:
        cpu_context.registers.cf = 0
        cpu_context.registers.af = 0
    cpu_context.registers.al &= 0xF

    logger.debug("Adjusted AX 0x%X -> 0x%X", orig_ax, cpu_context.registers.ax)


@opcode("adc")
@opcode("add")
def _add(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Handle both ADC and ADD here since the only difference is the flags.
    """
    operands = instruction.operands
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    result = opvalue1 + opvalue2
    if instruction.mnem == "adc":
        result += cpu_context.registers.cf
    width = get_max_operand_size(operands)

    mask = utils.get_mask(width)
    cpu_context.registers.cf = int(result > mask)
    cpu_context.registers.af = int((opvalue1 ^ opvalue2 ^ result) & 0x10)
    cpu_context.registers.zf = int(result & mask == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = int(not (-(mask // 2) <= result < (mask // 2)))
    cpu_context.registers.pf = get_parity(result)
    if cpu_context.emulator.branch_tracking:
        cpu_context.jcccontext.update_flag_opnds(["cf", "af", "zf", "sf", "of", "pf"], operands)

    logger.debug("0x%X + 0x%X = 0x%X", opvalue1, opvalue2, result)
    operands[0].value = result & mask


@opcode
def AND(cpu_context: ProcessorContext, instruction: Instruction):
    """ AND logic operator """
    operands = instruction.operands
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    result = opvalue1 & opvalue2
    width = get_max_operand_size(operands)

    cpu_context.registers.cf = 0
    cpu_context.registers.zf = int(result == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = 0
    cpu_context.registers.pf = get_parity(result)
    if cpu_context.emulator.branch_tracking:
        cpu_context.jcccontext.update_flag_opnds(["cf", "zf", "sf", "of", "pf"], operands)

    logger.debug("0x%X & 0x%X = 0x%X", opvalue1, opvalue2, result)
    operands[0].value = result


@opcode
def BSWAP(cpu_context: ProcessorContext, instruction: Instruction):
    """ byte Swap """
    operands = instruction.operands
    opvalue1 = operands[0].value
    width = operands[0].width
    result = swap_bytes(opvalue1, width)
    logger.debug("0x%X -> 0x%X", opvalue1, result)
    operands[0].value = result


@opcode
def CALL(cpu_context: ProcessorContext, instruction: Instruction):
    """
    CALL function

    Attempt to determine the number of arguments passed to the function which are purged on return
    """
    operands = instruction.operands
    # Function pointer can be a memory reference or immediate.
    func_ea = operands[0].addr or operands[0].value

    logger.debug("call %s", hex(func_ea))

    # If a valid function pointer, collect call history and emulate effects.
    if operands[0].is_func_ptr:
        # Using signature in order to get demangled name.
        try:
            func_name = cpu_context.emulator.disassembler.get_function_signature(func_ea).name
            logger.debug("call %s", func_name)
        except dragodis.NotExistError:
            logger.warning("Invalid function signature at %s", hex(func_ea))
            return

        # Push return address on the stack and set the ip to the function's start address.
        cpu_context.sp -= cpu_context.byteness
        ret_addr = instruction.next_ip
        cpu_context.memory.write(cpu_context.sp, ret_addr.to_bytes(cpu_context.byteness, cpu_context.byteorder))

        cpu_context._execute_call(func_ea, func_name, instruction.ip)

        # Pop return address from the stack, set ip to return address.
        cpu_context.sp += cpu_context.byteness
        cpu_context.ip = ret_addr
    else:
        logger.debug("Invalid function")

    if cpu_context.bitness == 64:
        return

    # Cleanup the stack based on the stack delta reported by the disassembler.
    cpu_context.sp += instruction._insn.stack_delta


@opcode
def CDQ(cpu_context: ProcessorContext, instruction: Instruction):
    """ Convert DWORD to QWORD with sign extension """
    if cpu_context.registers.eax >> 31:
        result = 0xFFFFFFFF
    else:
        result = 0x0

    logger.debug("Setting register EDX to 0x%X", result)
    cpu_context.registers.edx = result


@opcode
def CLC(cpu_context: ProcessorContext, instruction: Instruction):
    """ Clear Carry Flag """
    cpu_context.registers.cf = 0


@opcode
def CLD(cpu_context: ProcessorContext, instruction: Instruction):
    """ Clear Direction Flag """
    cpu_context.registers.df = 0


@opcode
def CMC(cpu_context: ProcessorContext, instruction: Instruction):
    """ Complement Carry Flag """
    cpu_context.registers.cf = int(not cpu_context.registers.cf)


@opcode
def CMP(cpu_context: ProcessorContext, instruction: Instruction):
    """ Compare to values """
    operands = instruction.operands
    width = get_min_operand_size(operands)
    mask = utils.get_mask(width)
    opvalue1 = operands[0].value & mask
    opvalue2 = operands[1].value & mask
    result = opvalue1 - opvalue2

    cpu_context.registers.cf = int((opvalue1 & mask) < (opvalue2 & mask))
    cpu_context.registers.af = int((opvalue1 ^ opvalue2 ^ result) & 0x10)
    cpu_context.registers.zf = int(result & mask == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = int(not (-(mask // 2) <= result < (mask // 2)))
    cpu_context.registers.pf = get_parity(result)
    if cpu_context.emulator.branch_tracking:
        cpu_context.jcccontext.update_flag_opnds(["cf", "af", "zf", "sf", "of", "pf"], operands)

    logger.debug("0x%X <-> 0x%X = 0x%X", opvalue1, opvalue2, result)


@opcode
def CMPS(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Nothing really to do for CMPS
    """
    pass


@opcode
def CMPSB(cpu_context: ProcessorContext, instruction: Instruction):
    """
    TODO: Does this really need to be implemented for our purposes???
    """
    pass


@opcode
def CMPSW(cpu_context: ProcessorContext, instruction: Instruction):
    """
    TODO: Does this really need to be implemented for our purposes???
    """
    pass


@opcode
def CMPSD(cpu_context: ProcessorContext, instruction: Instruction):
    """
    TODO: Does this really need to be implemented for our purposes???
    """
    pass


@opcode
def CQO(cpu_context: ProcessorContext, instruction: Instruction):
    """ Convert QWORD to DQWORD with sign extension """
    # Only works in 64-bit mode
    if cpu_context.bitness != 64:
        logger.debug("Opcode only available for 64-bit mode.")
        return

    if cpu_context.registers.rax >> 63:
        result = 0xFFFFFFFFFFFFFFFF
    else:
        result = 0x0

    logger.debug("Setting register RDX to 0x%X", result)
    cpu_context.registers.rdx = result


@opcode
def CVTDQ2PD(cpu_context: ProcessorContext, instruction: Instruction):
    """ Convert Packed Doubleword Integers to Packed Double-Precision Floating-Point Values """
    operands = instruction.operands
    opvalue2 = operands[1].value
    dword0 = opvalue2 & 0xFFFFFFFF
    dword1 = (opvalue2 & 0xFFFFFFFF) >> 32
    dpfp0 = utils.float_to_int(dword0)
    dpfp1 = utils.float_to_int(dword1)
    result = (dpfp1 << 64) | dpfp0
    logger.debug("0x%X -> 0x%X, 0x%X -> 0x%X --> 0x%X", dword0, dpfp0, dword1, dpfp1, result)
    operands[0].value = result


@opcode
def CVTSI2SD(cpu_context: ProcessorContext, instruction: Instruction):
    """ Convert Doubleword Int to Scalar Double-Precision Floating-Point """
    operands = instruction.operands
    opvalue2 = operands[1].value
    result = utils.float_to_int(opvalue2)
    logger.debug("int 0x%X -> float equivalent 0x%X", opvalue2, result)
    operands[0].value = result


@opcode
def CVTTSD2SI(cpu_context: ProcessorContext, instruction: Instruction):
    """ Convert with Truncation Scalar Double-Precision Floating-Point Value to Signed Integer """
    operands = instruction.operands
    opvalue2 = operands[1].value
    # width = operands[0].width
    result = int(utils.int_to_float(opvalue2))
    logger.debug("float 0x%X -> int equivalent 0x%X", opvalue2, result)
    operands[0].value = result


@opcode
def CWD(cpu_context: ProcessorContext, instruction: Instruction):
    """ Convert WORD to DWORD with sign extension """
    if cpu_context.registers.ax >> 15:
        result = 0xFFFF
    else:
        result = 0x0

    logger.debug("Setting register DX to 0x%X", result)
    cpu_context.registers.dx = result


@opcode
def DEC(cpu_context: ProcessorContext, instruction: Instruction):
    """ Decrement """
    operands = instruction.operands
    opvalue1 = operands[0].value
    width = operands[0].width
    mask = utils.get_mask(width)
    result = opvalue1 - 1

    cpu_context.registers.af = int(result & 0x0F == 0x0F)
    cpu_context.registers.zf = int(result & mask == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = int(utils.sign_bit(opvalue1, width) and not utils.sign_bit(result, width))
    cpu_context.registers.pf = get_parity(result)
    if cpu_context.emulator.branch_tracking:
        cpu_context.jcccontext.update_flag_opnds(["af", "zf", "sf", "of", "pf"], operands)

    logger.debug("0x%X - 1 = 0x%X", opvalue1, result)
    operands[0].value = result & mask


@opcode("div")
@opcode("idiv")
def _div(cpu_context: ProcessorContext, instruction: Instruction):
    signed = instruction.mnem == "idiv"
    RAX_REG_SIZE_MAP = {8: "rax", 4: "eax", 2: "ax", 1: "al"}
    RDX_REG_SIZE_MAP = {8: "rdx", 4: "edx", 2: "dx"}

    operands = instruction.operands
    # Need to obtain the width of the divisor, to determine the width of the dividend
    width = operands[-1].width
    divisor = operands[-1].value
    if signed:
        divisor = utils.signed(divisor, width)
    if divisor == 0:
        logger.debug("DIV / 0")
        return

    if width == 1:
        # When dividing by a 8-bit value, use AX
        dividend = cpu_context.registers.ax
        if signed:
            dividend = utils.signed(dividend, 2)
        result_reg = "al"
        remainder_reg = "ah"

    else:
        # When dividing by 16-bits -> combine DX:AX
        # When dividing by 32-bits -> combine EDX:EAX
        # When dividing by 64-bits -> combine RDX:RAX
        rax_str = RAX_REG_SIZE_MAP[width]
        rdx_str = RDX_REG_SIZE_MAP[width]
        dividend = (cpu_context.registers[rdx_str] << (width * 8)) | cpu_context.registers[rax_str]
        if signed:
            dividend = utils.signed(dividend, width * 2)
        result_reg = rax_str
        remainder_reg = rdx_str

    result = (dividend // divisor) & utils.get_mask(width)
    remainder = (dividend % divisor) & utils.get_mask(width)
    logger.debug("0x%X / 0x%X = 0x%X R 0x%X", dividend, divisor, result, remainder)

    cpu_context.registers[result_reg] = result
    cpu_context.registers[remainder_reg] = remainder


@opcode
def DIVSD(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Divide Scalar Double-Precision Floating-Point Value

    op1 / op2 -> op1
    """
    operands = instruction.operands
    dividend = utils.int_to_float(operands[0].value)
    divisor = utils.int_to_float(operands[1].value)
    # Because there is no guarantee that the registers/memory have been properly initialized, ignore DIV / 0 errors.
    if divisor == 0:
        # Log DIV / 0 error
        logger.debug("DIV / 0")
        return

    result = dividend / divisor
    logger.debug("%f / %f = %f", dividend, divisor, result)
    result = utils.float_to_int(result)
    operands[0].value = result


def _mul(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Handle MUL instruction and 1-operand IMUL instruction as the same.
    """
    RAX_REG_SIZE_MAP = {8: "rax", 4: "eax", 2: "ax", 1: "al"}
    RDX_REG_SIZE_MAP = {8: "rdx", 4: "edx", 2: "dx"}

    dx_reg = None
    dx_result = None
    operands = instruction.operands
    width = get_max_operand_size(operands)
    mask = utils.get_mask(width)
    multiplier1 = cpu_context.registers[RAX_REG_SIZE_MAP[width]]  # implied operand
    # Pull right-most operand to handle both 1 and 2 operand instructions
    multiplier2 = operands[-1].value
    result = multiplier1 * multiplier2
    flags = ["cf", "of"]

    if width == 1:
        ax_reg = RAX_REG_SIZE_MAP[2]
        ax_result = result
        if instruction.mnem == "mul":
            cpu_context.registers.cf = 0
            cpu_context.registers.of = 0
    else:
        ax_reg = RAX_REG_SIZE_MAP[width]
        dx_reg = RDX_REG_SIZE_MAP[width]
        dx_result = (result & (utils.get_mask(width) << (width * 8))) >> (width * 8)
        ax_result = result & utils.get_mask(width)
        if instruction.mnem == "mul":
            if result >> (width * 8):
                cpu_context.registers.cf = 1
                cpu_context.registers.of = 1
            else:
                cpu_context.registers.cf = 0
                cpu_context.registers.of = 0

    if instruction.mnem == "imul":
        cpu_context.registers.cf = int(
            not (
                (not utils.sign_bit(multiplier1, width) and multiplier2 & mask == 0)
                or (utils.sign_bit(multiplier1, width) and multiplier2 & mask == mask)
            )
        )
        cpu_context.registers.of = cpu_context.registers.cf
        cpu_context.registers.zf = int(multiplier1 & mask == 0)
        cpu_context.registers.sf = utils.sign_bit(multiplier1, width)
        cpu_context.registers.pf = get_parity(multiplier1)
        flags.extend(["zf", "sf", "pf"])

    if cpu_context.emulator.branch_tracking:
        cpu_context.jcccontext.update_flag_opnds(flags, operands)
    logger.debug(
        "0x%X * 0x%X = 0x%X || EAX -> 0x%X || EDX -> %s",
            multiplier1, multiplier2, result, ax_result, "0x%X" % dx_result if dx_reg else ""
        )

    cpu_context.registers[ax_reg] = ax_result
    if dx_reg:
        cpu_context.registers[dx_reg] = dx_result


# TODO: Clean up mul, imul, and _mul
@opcode
def IMUL(cpu_context: ProcessorContext, instruction: Instruction):
    """ Signed Multiplication

    ; Single operand form
    imul    ecx     ; Signed multiply the value in ecx with the value in eax (et.al)

    ; Two operand form
    imul    edi, edx    ; Signed multiply the destination operand (op 0) with the source operand (op 1)

    ; Three operand form
    imul    eax, edi, 5 ; Signed multiple source operand (op 1) with the immediate value (op 2) and store in
                        ; the destination operand (op 0)

    """
    operands = instruction.operands
    width = get_max_operand_size(operands)
    insn_data = instruction.data
    # Check for REX_W indicating a long instruction, REX is 0x40 with the W bit (0x08) set
    opcode_byte = insn_data[1] if insn_data[0] == 0x48 else insn_data[0]
    # F6/F7 are 8 and 16/32 bit IMUL without truncation so just use _mul
    if opcode_byte in (0xF6, 0xF7):
        _mul(cpu_context, instruction)
        return

    multiplier1 = operands[-2].value
    multiplier2 = operands[-1].value

    mask = utils.get_mask(width)
    result = multiplier1 * multiplier2

    cpu_context.registers.cf = int(
        not (
            (not utils.sign_bit(multiplier1, width) and multiplier2 & mask == 0)
            or (utils.sign_bit(multiplier1, width) and multiplier2 & mask == mask)
        )
    )
    cpu_context.registers.of = cpu_context.registers.cf
    cpu_context.registers.zf = int(multiplier1 & mask == 0)
    cpu_context.registers.sf = utils.sign_bit(multiplier1, width)
    cpu_context.registers.pf = get_parity(multiplier1)
    if cpu_context.emulator.branch_tracking:
        cpu_context.jcccontext.update_flag_opnds(["cf", "zf", "sf", "of", "pf"], operands)

    logger.debug("0x%X * 0x%X = 0x%X", multiplier1, multiplier2, result)
    operands[0].value = result


@opcode
def INC(cpu_context: ProcessorContext, instruction: Instruction):
    """ Increment """
    operands = instruction.operands
    opvalue1 = operands[0].value

    result = opvalue1 + 1
    width = operands[0].width
    mask = utils.get_mask(width)

    logger.debug("0x%X + 1 = 0x%X", opvalue1, result)
    operands[0].value = result

    cpu_context.registers.af = int(result & 0x0F == 0)
    cpu_context.registers.zf = int(result & mask == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = int(not utils.sign_bit(opvalue1, width) and utils.sign_bit(result, width))
    cpu_context.registers.pf = get_parity(result)
    if cpu_context.emulator.branch_tracking:
        cpu_context.jcccontext.update_flag_opnds(["af", "zf", "sf", "of", "pf"], operands)


@opcode
def JMP(cpu_context: ProcessorContext, instruction: Instruction):
    """ Unconditional jump """
    operands = instruction.operands
    jump_target = operands[0].value
    cpu_context.ip = jump_target


# TODO: Not all Jcc instructions are implemented here.
# TODO: Currently, these Jcc instructions assume that the instruction that modified the flags had 2 opcodes.
#   This is not always the case, (e.g. 'inc'). Also, we should probably figure out a way to move the
#   bulk of the code into a helper function.

# For the following jump instructions, the logic is basically the same
#   1. Get all the CodeRefs from the current IP (should only ever be 2)
#   2. Remove the EA that is the target of the Jcc instruction to we know where the non-jump target is
#   3. Determine the location where our condition takes us and set condition_target_ea to that
#   4. Set the value for the alternate path
# Note that since we aren't currently handling instructions which may cause conditional jumps, we need to
# determine if we have test_opnds and abort fixing the context if we don't.
@opcode("ja")
@opcode("jnbe")
def JA_JNBE(cpu_context: ProcessorContext, instruction: Instruction):
    """ Jump Above (CF=0 && ZF=0) """
    jump_target = instruction.operands[0].value
    jump = cpu_context.registers.cf == 0 and cpu_context.registers.zf == 0
    if jump:
        cpu_context.ip = jump_target

    if not cpu_context.emulator.branch_tracking:
        return

    next_inst = instruction._insn.line.next.address

    # Set the location where the condition would take use based on our emulation and the value for the alt branch
    test_operands = cpu_context.jcccontext.get_flag_opnds(["cf", "zf"])
    if len(test_operands) != 2:
        return

    operand0, operand1 = test_operands[:2]
    cpu_context.jcccontext.alt_branch_data_dst = operand0
    if jump:
        # opnd0 > opnd1 on this branch.  Set the alternate branch value opnd0 <= opnd1
        cpu_context.jcccontext.condition_target_ea = jump_target
        cpu_context.jcccontext.alt_branch_data = operand1.value - 1
    else:
        # opnd0 <= opnd1 on this branch. Set the alternate branch value opnd0 > opnd1
        cpu_context.jcccontext.condition_target_ea = next_inst
        cpu_context.jcccontext.alt_branch_data = operand1.value + 1

    logger.debug(
        "Primary branch 0x%X, using value 0x%X for alternate branch",
            cpu_context.jcccontext.condition_target_ea, cpu_context.jcccontext.alt_branch_data
        )

@opcode("jae")
@opcode("jnb")
@opcode("jnc")
def JAE_JNB(cpu_context: ProcessorContext, instruction: Instruction):
    """ Jump Above or Equal / Jump Not Below / Jump Not Carry (CF=0) """
    jump_target = instruction.operands[0].value
    jump = cpu_context.registers.cf == 0
    if jump:
        cpu_context.ip = jump_target

    if not cpu_context.emulator.branch_tracking:
        return

    next_inst = instruction._insn.line.next.address

    # Set the location where the condition would take use based on our emulation
    test_operands = cpu_context.jcccontext.get_flag_opnds(["cf"])
    if len(test_operands) != 2:
        return

    operand0, operand1 = test_operands[:2]
    cpu_context.jcccontext.alt_branch_data_dst = operand0
    if jump:
        # opnd0 > opnd1 on this branch.  Set the alternate branch value opnd0 < opnd1
        cpu_context.jcccontext.condition_target_ea = jump_target
        cpu_context.jcccontext.alt_branch_data = operand1.value - 1
    else:
        # opnd0 < opnd1 on this branch. Set the alternate branch value opnd0 >= opnd1
        cpu_context.jcccontext.condition_target_ea = next_inst
        cpu_context.jcccontext.alt_branch_data = operand1.value + 1

    logger.debug(
        "Primary branch 0x%X, using value 0x%X for alternate branch",
            cpu_context.jcccontext.condition_target_ea, cpu_context.jcccontext.alt_branch_data
        )


@opcode("jb")
@opcode("jc")
@opcode("jnae")
def JB_JNAE(cpu_context: ProcessorContext, instruction: Instruction):
    """ Jump Below / Jump Carry / Jump Not Above or Equal (CF=1) """
    jump_target = instruction.operands[0].value
    jump = cpu_context.registers.cf
    if jump:
        cpu_context.ip = jump_target

    if not cpu_context.emulator.branch_tracking:
        return

    # Don't know the data for either path specifically, but inferences can be made that the jump target will contain
    # a value in the first operand that is less than the second operand of the compare operation.
    next_inst = instruction._insn.line.next.address

    # Set the location where the condition would take use based on our emulation
    test_operands = cpu_context.jcccontext.get_flag_opnds(["cf"])
    if len(test_operands) != 2:
        return

    operand0, operand1 = test_operands[:2]
    cpu_context.jcccontext.alt_branch_data_dst = operand0
    if jump:
        # opnd0 < opnd1 on this branch.  Set the alternate branch value opnd0 >= opnd1
        cpu_context.jcccontext.condition_target_ea = jump_target
        cpu_context.jcccontext.alt_branch_data = operand1.value + 1
    else:
        # opnd0 >= opnd1 on this branch. Set the alternate branch value opnd0 < opnd1
        cpu_context.jcccontext.condition_target_ea = next_inst
        cpu_context.jcccontext.alt_branch_data = operand1.value - 1

    logger.debug(
        "Primary branch 0x%X, using value 0x%X for alternate branch",
            cpu_context.jcccontext.condition_target_ea, cpu_context.jcccontext.alt_branch_data
        )


@opcode("jbe")
@opcode("jna")
def JBE_JNA(cpu_context: ProcessorContext, instruction: Instruction):
    """ Jump Below or Equal / Jump Not Above (CF=1 || ZF=1) """
    jump_target = instruction.operands[0].value
    jump = cpu_context.registers.cf or cpu_context.registers.zf
    if jump:
        cpu_context.ip = jump_target

    if not cpu_context.emulator.branch_tracking:
        return

    # Don't know the data for either path specifically, but inferences can be made that the jump target will contain
    # a value in the first operand that is less than or equal to the second operand of the compare operation.
    next_inst = instruction._insn.line.next.address

    # Set the location where the condition would take use based on our emulation
    test_operands = cpu_context.jcccontext.get_flag_opnds(["cf", "zf"])
    if len(test_operands) != 2:
        return

    operand0, operand1 = test_operands[:2]
    cpu_context.jcccontext.alt_branch_data_dst = operand0
    if jump:
        # opnd0 <= opnd1 on this branch.  Set the alternate branch value opnd0 > opnd1
        cpu_context.jcccontext.condition_target_ea = jump_target
        cpu_context.jcccontext.alt_branch_data = operand1.value + 1
    else:
        # opnd0 > opnd1 on this branch.  Set the alternate branch value opnd0 <= opnd1
        cpu_context.jcccontext.condition_target_ea = next_inst
        cpu_context.jcccontext.alt_branch_data = operand1.value - 1

    logger.debug(
        "Primary branch 0x%X, using value 0x%X for alternate branch",
            cpu_context.jcccontext.condition_target_ea, cpu_context.jcccontext.alt_branch_data
        )


@opcode("je")
@opcode("jz")
def JE_JZ(cpu_context: ProcessorContext, instruction: Instruction):
    """ Jump Equal / Jump Zero (ZF=1) """
    jump_target = instruction.operands[0].value
    jump = cpu_context.registers.zf
    if jump:
        cpu_context.ip = jump_target

    if not cpu_context.emulator.branch_tracking:
        return

    # Jump target contains the known data which is either 0 or the value of the second operand of the compare operation.
    next_inst = instruction._insn.line.next.address

    # Set the location where the condition would take use based on our emulation
    test_operands = cpu_context.jcccontext.get_flag_opnds(["zf"])
    if len(test_operands) != 2:
        return

    operand0, operand1 = test_operands[:2]
    cpu_context.jcccontext.alt_branch_data_dst = operand0
    # There is additional logic that must be conducted for this jump.  If the src and dst operands are the same, then
    # the check was likely to determine if the value was 0 or not 0.  Else, the check was determining if src and dst
    # were equal.
    if jump:
        # Indicates emulation likely produced two operands which were equal, or an operand which was 0, so set the
        # alternate branch to equal the value of operand 1 + 1.  Or the comparison was to check if an operand was 0.
        # eg:  cmp    eax, 0x3D     ; eax = 0x3D
        #      cmp    eax, 0x00     ; eax = 0
        #      test   rax, rax      ; rax = 0
        cpu_context.jcccontext.condition_target_ea = jump_target
        # There is no need to check if the operands are the same here.  If the check was for equality, then adding
        # 1 to the second operand will make the check not equal.  If the check was for 0 (both operands were the same),
        # simply adding 1 to 0 will make the test fail.
        cpu_context.jcccontext.alt_branch_data = operand1.value + 1
    else:
        # Indicates emulation likely produced two operands which were not equal, or an operand which was not 0, so set
        # the alternate branch to equal the value of operand 1.  Or the comparison was to check if an operand was 0.
        # eg:  cmp    eax, 0x3D     ; eax = 0
        #      cmp    eax, 0x00     ; eax = 7
        #      test   rax, rax      ; rax = 10
        cpu_context.jcccontext.condition_target_ea = next_inst
        # Need to determine if both operands are the same.  If they are, then the check was actually to determine
        # if the value was 0.  In this case, to make the test succeed, just set the value to 0.  Otherwise the test
        # was to compare two different operands for equality, so just set the first operand to the value of the second.
        if operand0.text == operand1.text:
            cpu_context.jcccontext.alt_branch_data = 0
        else:
            cpu_context.jcccontext.alt_branch_data = operand1.value

    logger.debug(
        "Primary branch 0x%X, using value 0x%X for alternate branch",
            cpu_context.jcccontext.condition_target_ea, cpu_context.jcccontext.alt_branch_data
        )


@opcode("jg")
@opcode("jnle")
def JG_JNLE(cpu_context: ProcessorContext, instruction: Instruction):
    """ Jump Greater / Jump Not Less or Equal (ZF=0 && SF=OF) """
    jump_target = instruction.operands[0].value
    jump = cpu_context.registers.zf == 0 and cpu_context.registers.sf == cpu_context.registers.of
    if jump:
        cpu_context.ip = jump_target

    if not cpu_context.emulator.branch_tracking:
        return

    next_inst = instruction._insn.line.next.address

    # Don't know the data for either path specifically, but inferences can be made that the jump target will contain
    # a value in the first operand that is larger than the second operand of the compare operation.

    # Set the location where the condition we would take used based on our emulation
    test_operands = cpu_context.jcccontext.get_flag_opnds(["zf", "sf"])
    if len(test_operands) != 2:
        return

    operand0, operand1 = test_operands[:2]
    cpu_context.jcccontext.alt_branch_data_dst = operand0
    if jump:
        # opnd0 > opnd1 on this branch.  Set alternate branch value opnd0 <= opnd1
        cpu_context.jcccontext.condition_target_ea = jump_target
        cpu_context.jcccontext.alt_branch_data = operand1.value - 1
    else:
        # opnd0 <= opnd1on this branch.  Set alternate branch value opnd0 > opnd1
        cpu_context.jcccontext.condition_target_ea = next_inst
        cpu_context.jcccontext.alt_branch_data = operand1.value + 1

    logger.debug(
        "Primary branch 0x%X, using value 0x%X for alternate branch",
            cpu_context.jcccontext.condition_target_ea, cpu_context.jcccontext.alt_branch_data
        )


@opcode("jge")
@opcode("jnl")
def JGE_JNL(cpu_context: ProcessorContext, instruction: Instruction):
    """ Jump Greater or Equal (SF=OF) """
    jump_target = instruction.operands[0].value
    jump = cpu_context.registers.sf == cpu_context.registers.of
    if jump:
        cpu_context.ip = jump_target

    if not cpu_context.emulator.branch_tracking:
        return

    # Don't know the data for either path specifically, but inferences can be made that the jump target will contain
    # a value in the first operand that is larger than or equal to the second operand of the compare operation.
    next_inst = instruction._insn.line.next.address

    # Set the location where the condition we would take used based on our emulation
    test_operands = cpu_context.jcccontext.get_flag_opnds(["sf", "of"])
    if len(test_operands) != 2:
        return

    operand0, operand1 = test_operands[:2]
    cpu_context.jcccontext.alt_branch_data_dst = operand0
    if jump:
        # opnd0 >= opnd1 on this branch. Set alternate branch value opnd0 < opnd1
        cpu_context.jcccontext.condition_target_ea = jump_target
        cpu_context.jcccontext.alt_branch_data = operand1.value - 1
    else:
        # opnd0 < opnd1 on this branch. Set alternate branch value opnd0 >= opnd1
        cpu_context.jcccontext.condition_target_ea = next_inst
        cpu_context.jcccontext.alt_branch_data = operand1.value + 1

    logger.debug(
        "Primary branch 0x%X, using value 0x%X for alternate branch",
            cpu_context.jcccontext.condition_target_ea, cpu_context.jcccontext.alt_branch_data
        )


@opcode("jl")
@opcode("jnge")
def JL_JNGE(cpu_context: ProcessorContext, instruction: Instruction):
    """ Jump Less (SF!=OF) """
    jump_target = instruction.operands[0].value
    jump = cpu_context.registers.sf != cpu_context.registers.of
    if jump:
        cpu_context.ip = jump_target

    if not cpu_context.emulator.branch_tracking:
        return

    # Don't know the data for either path specifically, but inferences can be made that the jump target will contain
    # a value in the first operand that is less than the second operand of the compare operation.
    next_inst = instruction._insn.line.next.address

    # Set the location where the condition we would take used based on our emulation
    test_operands = cpu_context.jcccontext.get_flag_opnds(["sf", "of"])
    if len(test_operands) != 2:
        return

    operand0, operand1 = test_operands[:2]
    cpu_context.jcccontext.alt_branch_data_dst = operand0
    if jump:
        # opnd0 < opnd1 on this branch.  Set alternate branch value opnd0 >= opnd1
        cpu_context.jcccontext.condition_target_ea = jump_target
        cpu_context.jcccontext.alt_branch_data = operand1.value + 1
    else:
        # opnd1 >= opnd1 on this branch.  Set alternate branch value opnd0 < opnd1
        cpu_context.jcccontext.condition_target_ea = next_inst
        cpu_context.jcccontext.alt_branch_data = operand1.value - 1

    logger.debug(
        "Primary branch 0x%X, using value 0x%X for alternate branch",
            cpu_context.jcccontext.condition_target_ea, cpu_context.jcccontext.alt_branch_data
        )


@opcode("jle")
@opcode("jng")
def JLE_JNG(cpu_context: ProcessorContext, instruction: Instruction):
    """ Jump Less or Equal (ZF=1 || SF!=OF) """
    jump_target = instruction.operands[0].value
    jump = cpu_context.registers.zf or cpu_context.registers.sf != cpu_context.registers.of
    if jump:
        cpu_context.ip = jump_target

    if not cpu_context.emulator.branch_tracking:
        return

   # Don't know the data for either path specifically, but inferences can be made that the jump target will contain
    # a value in the first operand that is less than or equal to the second operand of the compare operation.
    next_inst = instruction._insn.line.next.address

    # Set the location where the condition we would take used based on our emulation
    test_operands = cpu_context.jcccontext.get_flag_opnds(["zf", "sf", "of"])
    if len(test_operands) != 2:
        return

    operand0, operand1 = test_operands[:2]
    cpu_context.jcccontext.alt_branch_data_dst = operand0
    if jump:
        # opnd0 <= opnd2 on this branch.  Set alternate branch value opnd1 > opnd2
        cpu_context.jcccontext.condition_target_ea = jump_target
        cpu_context.jcccontext.alt_branch_data = operand1.value + 1
    else:
        # opnd0 > opnd2 on this branch.  Set alternate branch value opnd0 <= opnd2
        cpu_context.jcccontext.condition_target_ea = next_inst
        cpu_context.jcccontext.alt_branch_data = operand1.value - 1

    logger.debug(
        "Primary branch 0x%X, using value 0x%X for alternate branch",
            cpu_context.jcccontext.condition_target_ea, cpu_context.jcccontext.alt_branch_data
        )


@opcode("jne")
@opcode("jnz")
def JNE_JNZ(cpu_context: ProcessorContext, instruction: Instruction):
    """ Jump Not Equal (ZF=0) """
    jump_target = instruction.operands[0].value
    jump = cpu_context.registers.zf == 0
    if jump:
        cpu_context.ip = jump_target

    if not cpu_context.emulator.branch_tracking:
        return

    next_inst = instruction._insn.line.next.address

    # Whatever the operation, it either set ZF or it didn't... Typically, the assumption can probably be made that
    # either the operands were equal such that a subtraction resulted in 0, or they weren't.
    # TODO: Does the compare instruction have an effect on which operand is the value to be used?

    ## Set the target for which to modify the context, it will be the only address left in code_refs
    test_operands = cpu_context.jcccontext.get_flag_opnds(["zf"])
    if len(test_operands) != 2:
        return

    operand0, operand1 = test_operands[:2]
    cpu_context.jcccontext.alt_branch_data_dst = operand0
    # There is additional logic that must be conducted for this jump.  If the src and dst operands are the same, then
    # the check was likely determine if the value was 0 or not 0.  Else, the check was determining if src and dst were
    # not equal.
    if jump:
        # Indicates emulation likely produced two operands which were not equal, or an operand which was not 0, so set
        # the alternate branch to equal the value of operand 1 (which would also be 0).  Or the comparison was to check
        # if an operand was 0.
        # eg:  cmp    eax, 0x3D     ; eax = 0
        #      cmp    eax, 0x00     ; eax = 7
        #      test   rax, rax      ; rax = 10
        cpu_context.jcccontext.condition_target_ea = jump_target
        # Need to determine if the operands are the same operand (indicating a test for 0).  If the operands are the
        # same operand, then set the value to 0.  If the operands were not the same operand, the test was for
        # equality, so set the values to the same value.
        if operand0.text == operand1.text:
            cpu_context.jcccontext.alt_branch_data = 0
        else:
            cpu_context.jcccontext.alt_branch_data = operand1.value
    else:
        # Indicates emulation likely produced two operands which were equal, or an operand was 0, so set the
        # alternate branch to the value of operand 1 + 1.  Or the comparison was to check if an operand was 0.
        # eg:  cmp    eax, 0x00     ; eax = 0
        #      cmp    eax, 0x3D     ; eax = 0x3D
        #      test   rax, rax      ; rax = 0
        cpu_context.jcccontext.condition_target_ea = next_inst
        # There is no need to check if the operands were the same.  If the operands were the same, the test was to
        # determine if the operand was 0, so just adding 1 to the operand will be enough to create the false condition.
        # If the operands weren't the same, the check was to determine if the operands were equal, so adding 1 to
        # the second operand will be enough to make the condition false.
        cpu_context.jcccontext.alt_branch_data = operand1.value + 1

    logger.debug(
        "Primary branch 0x%X, using value 0x%X for alternate branch",
            cpu_context.jcccontext.condition_target_ea, cpu_context.jcccontext.alt_branch_data
        )


@opcode
def JNO(cpu_context: ProcessorContext, instruction: Instruction):
    """ Jump Not Overflow (OF=0) """
    if not cpu_context.registers.of:
        cpu_context.ip = instruction.operands[0].value


@opcode("jnp")
@opcode("jpo")
def JNP_JPO(cpu_context: ProcessorContext, instruction: Instruction):
    """ Jump Not Parity (PF=0) """
    if not cpu_context.registers.pf:
        cpu_context.ip = instruction.operands[0].value


@opcode
def JNS(cpu_context: ProcessorContext, instruction: Instruction):
    """ Jump Not Sign (SF=0) """
    if not cpu_context.registers.sf:
        cpu_context.ip = instruction.operands[0].value


@opcode
def JO(cpu_context: ProcessorContext, instruction: Instruction):
    """ Jump Overflow (OF=1) """
    if cpu_context.registers.of:
        cpu_context.ip = instruction.operands[0].value


@opcode("jp")
@opcode("jpe")
def JP_JPE(cpu_context: ProcessorContext, instruction: Instruction):
    """ Jump Parity (PF=1) """
    if cpu_context.registers.pf:
        cpu_context.ip = instruction.operands[0].value


@opcode
def JS(cpu_context: ProcessorContext, instruction: Instruction):
    """ Jump Sign (SF=1) """
    if cpu_context.registers.sf:
        cpu_context.ip = instruction.operands[0].value


@opcode
def LEA(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Handle the LEA instruction.
    """
    operands = instruction.operands
    address = operands[1].addr
    logger.debug("Copy address 0x%X into %s", address, operands[0].text)
    operands[0].value = address


@opcode("mov")
@opcode("movzx")
@opcode("movapd")
@opcode("movaps")
@opcode("movdqa")
@opcode("movdqu")
@opcode("movupd")
@opcode("movups")
def _mov(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Handle the MOV, MOVZX, MOVA*, MOVD*, MOVU* instructions in the same manner.

    MOVZX is a zero extend, but this logic makes no real sense in python.

    NOTE: Since the widths are already taken into account when the operand values are retrieved
    or set, the logic for most mov* instructions are the same.
    """
    operands = instruction.operands
    opvalue2 = operands[1].value
    logger.debug("Copy 0x%X into %s", opvalue2, operands[0].text)
    operands[0].value = opvalue2


@opcode("movsx")
@opcode("movsxd")
def _movsx(cpu_context: ProcessorContext, instruction: Instruction):
    """ Move with Sign Extend """
    operands = instruction.operands
    opvalue2 = operands[1].value
    logger.debug("Sign-extend 0x%X into %s", opvalue2, operands[0].text)
    size = utils.sign_extend(opvalue2, operands[1].width, operands[0].width)
    operands[0].value = size


@opcode("movs")  # I don't believe IDA will ever use just "movs", but it's here just incase.
@opcode("movsb")
@opcode("movsw")
@opcode("movsd")
@opcode("movsq")
def movs(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Move Scalar Double-Precision Floating-Point Value
    OR
    Move Data from String to String
    """
    operands = instruction.operands
    # Scalar Double-Precision Floating-point move
    # Need to amend this to take into account the VEX and EVEX prefixes (not sure what they are at this point)
    # references:
    #   en.wikipedia.org/wiki/VEX_prefix
    #   en.wikipedia.org/wiki/EVEX_prefix
    #if instruction.mnem == "movsd" and len(operands) == 2:
    if instruction.data[0] == 0xF2:
        op1, op2 = operands
        data = op2.value
        if op1.type == OperandType.register:
            # When moving into an XMM register, the high 64 bits needs to remain untouched.
            data = (data & 0xFFFFFFFFFFFFFFFF0000000000000000) | data
        logger.debug("0x%X -> 0x%X", op2.value, data)
        op1.value = data

    # movs*
    else:
        if cpu_context.bitness == 16:
            src = "si"
            dst = "di"
        # 0x67 indicates 32-bit addressing
        elif cpu_context.bitness == 64 and instruction.data[0] != 0x67:
            src = "rsi"
            dst = "rdi"
        else:
            src = "esi"
            dst = "edi"
        # IDA sometimes provides a single "fake" operand to help determine the size.
        width = operands[0].width if operands else 4

        size = {"movs": width, "movsb": 1, "movsw": 2, "movsd": 4, "movsq": 8}[instruction.mnem]
        src_ptr = cpu_context.registers[src]
        dst_ptr = cpu_context.registers[dst]
        logger.debug("0x%X -> 0x%X", src_ptr, dst_ptr)
        cpu_context.memory.copy(src_ptr, dst_ptr, size)

        # update ESI/EDI registers
        if cpu_context.registers.df:
            cpu_context.registers[src] -= size
            cpu_context.registers[dst] -= size
        else:
            cpu_context.registers[src] += size
            cpu_context.registers[dst] += size


@opcode
def MOVD(cpu_context: ProcessorContext, instruction: Instruction):
    """ Move Dword """
    operands = instruction.operands
    opvalue2 = operands[1].value & 0xFFFFFFFF
    logger.debug("Copy 0x%X into %s", opvalue2, operands[0].text)
    operands[0].value = opvalue2


@opcode
def MOVQ(cpu_context: ProcessorContext, instruction: Instruction):
    """ Move Quadword """
    operands = instruction.operands
    opvalue2 = operands[1].value & 0xFFFFFFFFFFFFFFFF
    logger.debug("Copy 0x%X into %s", opvalue2, operands[0].text)
    operands[0].value = opvalue2


@opcode
def MUL(cpu_context: ProcessorContext, instruction: Instruction):
    """ Multiplication """
    _mul(cpu_context, instruction)


@opcode
def NEG(cpu_context: ProcessorContext, instruction: Instruction):
    """ Negate """
    operands = instruction.operands
    opvalue1 = operands[0].value
    result = -opvalue1
    width = operands[0].width
    mask = utils.get_mask(width)

    cpu_context.registers.cf = int(result & mask != 0)
    cpu_context.registers.af = int(result & 0x0F != 0)
    cpu_context.registers.zf = int(result & mask == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = int(utils.sign_bit(opvalue1, width) and not utils.sign_bit(result, width))
    if cpu_context.emulator.branch_tracking:
        cpu_context.jcccontext.update_flag_opnds(["cf", "af", "zf", "sf", "of"], operands)

    logger.debug("0x%X - 0x%X", opvalue1, result)
    operands[0].value = result


@opcode
def NOT(cpu_context: ProcessorContext, instruction: Instruction):
    """ NOT Logic Operator """
    operands = instruction.operands
    opvalue1 = operands[0].value
    result = ~opvalue1
    logger.debug("0x%X -> 0x%X", opvalue1, result)
    operands[0].value = result


@opcode
def OR(cpu_context: ProcessorContext, instruction: Instruction):
    """ OR Logic Operator """
    operands = instruction.operands
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)
    result = opvalue1 | opvalue2

    cpu_context.registers.cf = 0
    cpu_context.registers.zf = int(result == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = 0
    cpu_context.registers.pf = get_parity(result)
    if cpu_context.emulator.branch_tracking:
        cpu_context.jcccontext.update_flag_opnds(["cf", "zf", "sf", "of", "pf"], operands)

    logger.debug("0x%X | 0x%X = 0x%X", opvalue1, opvalue2, result)
    operands[0].value = result


@opcode
def POP(cpu_context: ProcessorContext, instruction: Instruction):
    """ POP stack value """
    operands = instruction.operands
    data = cpu_context.memory.read(cpu_context.sp, cpu_context.byteness)
    result = int.from_bytes(data, cpu_context.byteorder)
    cpu_context.sp += cpu_context.byteness
    logger.debug("Popped value 0x%X into %s", result, operands[0].text)
    operands[0].value = result


@opcode("popa")
@opcode("popad")
def POPA(cpu_context: ProcessorContext, instruction: Instruction):
    """
    POPA (valid only for x86)

    NOTE: This function will return None.  This is one instance where accessing the registers directly makes more
            sense.
    """
    # NOTE Some assemblers may force size based on operand size instead of mnem.
    # However, IDA should set the proper mnemonic for us.
    if instruction.mnem.endswith("d"):
        reg_order = ["edi", "esi", "ebp", "esp", "ebx", "edx", "ecx", "eax"]
    else:
        reg_order = ["di", "si", "bp", "sp", "bx", "dx", "cx", "ax"]

    for reg in reg_order:
        if reg not in ("esp", "sp"):
            # reg <- Pop()
            data = cpu_context.memory.read(cpu_context.registers.esp, cpu_context.byteness)
            val = int.from_bytes(data, cpu_context.byteorder)
            cpu_context.registers[reg] = val
            logger.debug("Popped value 0x%X into %s", val, reg)
        cpu_context.sp += cpu_context.byteness


@opcode("popf")
@opcode("popfd")
@opcode("popfq")
def POPF(cpu_context: ProcessorContext, instruction: Instruction):
    """ Pop FLAGS/EFLAGS register off the stack """
    data = cpu_context.memory.read(cpu_context.sp, cpu_context.byteness)
    flags = int.from_bytes(data, cpu_context.byteorder)
    cpu_context.sp += cpu_context.byteness

    logger.debug("Popped value 0x%X into flags register", flags)
    if cpu_context.bitness == 16:
        cpu_context.registers.flags = flags
    else:
        # Also works for RFLAGS. Since we don't support them, they are all zeros.
        cpu_context.registers.eflags = flags


@opcode
def PUSH(cpu_context: ProcessorContext, instruction: Instruction):
    """ PUSH """
    operand = instruction.operands[0]
    value = utils.unsigned(operand.value, operand.width)
    logger.debug("Pushing 0x%X onto stack", value)
    cpu_context.registers.rsp -= cpu_context.byteness
    cpu_context.memory.write(
        cpu_context.registers.esp,
        value.to_bytes(operand.width, cpu_context.byteorder)
    )


@opcode("pusha")
@opcode("pushad")
def PUSHA(cpu_context: ProcessorContext, instruction: Instruction):
    """ Push all general-purpose registers (valid only for x86) """
    # NOTE Some assemblers may force size based on operand size instead of mnem.
    # However, IDA should set the proper mnemonic for us.
    if instruction.mnem.endswith("d"):
        reg_order = ["eax", "ecx", "edx", "ebx", "esp", "ebp", "esi", "edi"]
        orig_esp = cpu_context.registers.esp
    else:
        reg_order = ["ax", "cx", "dx", "bx", "sp", "bp", "si", "di"]
        orig_esp = cpu_context.registers.sp

    for reg in reg_order:
        cpu_context.sp -= cpu_context.byteness
        pushed_value = orig_esp if reg in ("esp", "sp") else cpu_context.registers[reg]
        logger.debug("Pushing 0x%X onto stack", pushed_value)
        cpu_context.memory.write(
            cpu_context.registers.esp,
            pushed_value.to_bytes(cpu_context.byteness, cpu_context.byteorder),
        )


@opcode("pushf")
@opcode("pushfd")
@opcode("pushfq")
def PUSHF(cpu_context: ProcessorContext, instruction: Instruction):
    """ Push FLAGS/EFLAGS register onto the stack """
    if cpu_context.bitness == 16:
        flags = cpu_context.registers.flags
    else:
        # Also works for RFLAGS. Since we don't support them, they are all zeros.
        flags = cpu_context.registers.eflags

    # VM and RF flags are not copied.
    flags &= ~0x10000  # rf
    flags &= ~0x20000  # vm

    logger.debug("Pushing 0x%X onto the stack", flags)
    cpu_context.sp -= cpu_context.byteness
    cpu_context.memory.write(
        cpu_context.sp,
        flags.to_bytes(cpu_context.byteness, cpu_context.byteorder),
    )


@opcode
def RCR(cpu_context: ProcessorContext, instruction: Instruction):
    """ Rotate Carry Right """
    operands = instruction.operands
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)

    # Because we want to allow for 64-bit code, we'll use 0x3F as our mask.
    if width == 8:
        tempcount = opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)
    elif width in [1, 2, 4]:
        # Basically MOD by 9, 17, 33 when the width is 1 byte, 2 bytes or 4 bytes
        tempcount = (opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)) % ((width * 8) + 1)
    else:
        # This is undefined behavior
        return

    if opvalue2 == 1:
        cpu_context.registers.of = get_msb(opvalue1, width) ^ cpu_context.registers.cf

    while tempcount:
        tempcf = get_lsb(opvalue2)
        opvalue1 = (opvalue1 >> 1) + (cpu_context.registers.cf * 2 ** width)
        cpu_context.registers.cf = tempcf
        tempcount -= 1

    if cpu_context.emulator.branch_tracking:
        cpu_context.jcccontext.update_flag_opnds(["cf"], operands)
    logger.debug("Rotate 0x%X right by 0x%X -> 0x%X",
                 operands[0].value, opvalue2, opvalue1)
    operands[0].value = opvalue1


@opcode
def RCL(cpu_context: ProcessorContext, instruction: Instruction):
    """ Rotate Carry Left """
    operands = instruction.operands
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)

    # Because we want to allow for 64-bit code, we'll use 0x3F as our mask.
    if width == 8:
        tempcount = opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)
    elif width in [1, 2, 4]:
        # Basically MOD by 9, 17, 33 when width is 1 byte, 2 bytes or 4 bytes
        tempcount = (opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)) % ((width * 8) + 1)
    else:
        # This is undefined behavior
        return

    while tempcount:
        tempcf = get_msb(opvalue1, width)
        opvalue1 = (opvalue1 * 2) + cpu_context.registers.cf
        cpu_context.registers.cf = tempcf
        tempcount -= 1

    if opvalue2 == 1:
        cpu_context.registers.of = get_msb(opvalue1, width) ^ cpu_context.registers.cf

    if cpu_context.emulator.branch_tracking:
        cpu_context.jcccontext.update_flag_opnds(["cf", "of"], operands)
    logger.debug("Rotate 0x%X left by 0x%X -> 0x%X",
                 operands[0].value, opvalue2, opvalue1)
    operands[0].value = opvalue1


@opcode
def ROL(cpu_context: ProcessorContext, instruction: Instruction):
    """ Rotate Left """
    operands = instruction.operands
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)

    # Because we want to allow for 64-bit code, we'll use 0x3F as our mask.
    if width == 8:
        tempcount = opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)
    elif width in [1, 2, 4]:
        # Basically MOD by 8, 16, 32 when width is 1 byte, 2 bytes or 4 bytes
        tempcount = (opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)) % (width * 8)
    else:
        # This is undefined behavior
        return

    if tempcount > 0:
        mask = utils.get_mask(width)
        while tempcount:
            opvalue1 = (opvalue1 * 2) + get_msb(opvalue1, width)
            opvalue1 &= mask
            tempcount -= 1

        cpu_context.registers.cf = get_lsb(opvalue1)
        if opvalue2 == 1:
            cpu_context.registers.of = get_msb(opvalue1, width) ^ cpu_context.registers.cf

    if cpu_context.emulator.branch_tracking:
        cpu_context.jcccontext.update_flag_opnds(["cf", "of"], operands)
    logger.debug("Rotate 0x%X left by 0x%X -> 0x%X",
                 operands[0].value, opvalue2, opvalue1)
    operands[0].value = opvalue1


@opcode
def ROR(cpu_context: ProcessorContext, instruction: Instruction):
    """ Rotate Right """
    operands = instruction.operands
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)

    # Because we want to allow for 64-bit code, we'll use 0x3F as our mask.
    if width == 8:
        tempcount = opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)
    elif width in [1, 2, 4]:
        # Basically MOD by 8, 16, 32 when width is 1 byte, 2 bytes or 4 bytes
        tempcount = (opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)) % (width * 8)
    else:
        # This is undefined behavior
        return

    if tempcount > 0:
        while tempcount:
            tempcf = get_lsb(opvalue2)
            opvalue1 = (opvalue1 >> 1) + (tempcf * 2 ** width)
            tempcount -= 1

        cpu_context.registers.cf = get_msb(opvalue1, width)
        if opvalue2 == 1:
            cpu_context.registers.of = get_msb(opvalue1, width) ^ (get_msb(opvalue1, width) - 1)

    if cpu_context.emulator.branch_tracking:
        cpu_context.jcccontext.update_flag_opnds(["cf", "of"], operands)
    logger.debug("Rotate 0x%X right by 0x%X -> 0x%X",
                 operands[0].value, opvalue2, opvalue1)
    operands[0].value = opvalue1


@opcode("sal")
@opcode("shl")
def sal_shl(cpu_context: ProcessorContext, instruction: Instruction):
    """ Shift Arithmetic Left """
    operands = instruction.operands
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)
    result = opvalue1

    if opvalue2:
        # 0x3F Because we want to allow for 64-bit code
        tempcount = opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)
        mask = utils.get_mask(width)
        while tempcount:
            cpu_context.registers.cf = get_msb(result, width)
            result *= 2
            result &= mask
            tempcount -= 1

        result &= utils.get_mask(width)

        bit_count = width * 8
        if opvalue2 <= bit_count:
            cpu_context.registers.cf = (opvalue1 >> (bit_count - opvalue2)) & 0x01
        else:
            cpu_context.registers.cf = 0
        cpu_context.registers.sf = utils.sign_bit(result, width)
        if opvalue2 == 1:
            cpu_context.registers.of = utils.sign_bit(opvalue1 ^ result, width)
        else:
            cpu_context.registers.of = utils.sign_bit((opvalue1 << (opvalue2 - 1)) ^ result, width)

    if cpu_context.emulator.branch_tracking:
        cpu_context.jcccontext.update_flag_opnds(["cf", "of"], operands)
    logger.debug("Shift 0x%X left by 0x%X -> 0x%X", opvalue1, opvalue2, result)
    operands[0].value = result


@opcode
def SAR(cpu_context: ProcessorContext, instruction: Instruction):
    """ Shift Arithmetic Right """
    operands = instruction.operands
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)
    result = opvalue1
    if opvalue2:
        # 0x3F Because we want to allow for 64-bit code
        tempcount = opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)
        msb = get_msb(opvalue1, cpu_context.byteness)
        while tempcount:
            cpu_context.registers.cf = get_lsb(result)
            result = result >> 1
            tempcount -= 1

        bit_count = width * 8
        if opvalue2 < bit_count:
            cpu_context.registers.cf = (opvalue1 >> (opvalue2 - 1)) & 0x01
        else:
            cpu_context.registers.cf = utils.sign_bit(opvalue1, width)
        cpu_context.registers.zf = int(result == 0)
        cpu_context.registers.sf = utils.sign_bit(result, width)
        cpu_context.registers.of = 0
        cpu_context.registers.pf = get_parity(result)

        result |= msb << cpu_context.bitness

    if cpu_context.emulator.branch_tracking:
        cpu_context.jcccontext.update_flag_opnds(["cf", "zf", "sf", "of", "pf"], operands)
    logger.debug("Shift 0x%X right by 0x%X -> 0x%X", opvalue1, opvalue2, result)
    operands[0].value = result


@opcode
def SBB(cpu_context: ProcessorContext, instruction: Instruction):
    """ Subtract with Borrow/Carry """
    operands = instruction.operands
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)
    mask = utils.get_mask(width)
    result = opvalue1 - (opvalue2 + cpu_context.registers.cf)

    cpu_context.registers.af = int((opvalue1 ^ opvalue2 ^ result) & 0x10)
    cpu_context.registers.zf = int(result == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = int(not (-(mask // 2) <= result < (mask // 2)))
    cpu_context.registers.pf = get_parity(result)
    if cpu_context.emulator.branch_tracking:
        cpu_context.jcccontext.update_flag_opnds(["af", "zf", "sf", "of", "pf"], operands)
    logger.debug("0x%X - 0x%X = 0x%X",
                 opvalue1, (opvalue2 + cpu_context.registers.cf), result)
    operands[0].value = result


@opcode("scas")
@opcode("scasb")
@opcode("scasw")
@opcode("scasd")
def scas(cpu_context: ProcessorContext, instruction: Instruction):
    """ Scan string """
    if cpu_context.bitness == 16:
        edi_reg = "di"
    else:
        edi_reg = "edi"

    mnem = instruction.mnem
    if mnem.endswith("b"):
        width = 1
    elif mnem.endswith("w"):
        width = 2
    elif mnem.endswith("d"):
        width = 4
    else:
        width = instruction.operands[0].width

    eax_reg = {1: "al", 2: "ax", 4: "eax"}[width]

    # Compare value in eax with value at memory location stored in edi.
    opvalue1 = cpu_context.registers[eax_reg]
    data = cpu_context.memory.read(cpu_context.registers[edi_reg], width)
    opvalue2 = int.from_bytes(data, cpu_context.byteorder)
    result = opvalue1 - opvalue2
    logger.debug("Scan compare 0x%X - 0x%X = 0x%X", opvalue1, opvalue2, result)

    mask = utils.get_mask(width)
    cpu_context.registers.cf = int((opvalue1 & mask) < (opvalue2 & mask))
    cpu_context.registers.af = int((opvalue1 ^ opvalue2 ^ result) & 0x10)
    cpu_context.registers.zf = int(result & mask == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = int(not (-(mask // 2) <= result < (mask // 2)))
    cpu_context.registers.pf = get_parity(result)
    # TODO: Can't branch track because no real operands.
    # if cpu_context.emulator.branch_tracking:
    #     cpu_context.jcccontext.update_flag_opnds(["cf", "af", "zf", "sf", "of", "pf"], operands)

    # Update or decrement edi
    if cpu_context.registers.df:
        cpu_context.registers[edi_reg] -= width
    else:
        cpu_context.registers[edi_reg] += width


@opcode
def SETNA(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Not Above """
    operands = instruction.operands
    result = int(cpu_context.registers.zf or cpu_context.registers.cf)
    logger.debug("Setting %s to 0x%X", operands[0].text, result)
    operands[0].value = result


@opcode
def SETLE(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Less than or Equal """
    operands = instruction.operands
    result = int(cpu_context.registers.zf or (cpu_context.registers.sf != cpu_context.registers.of))
    logger.debug("Setting %s to 0x%X", operands[0].text, result)
    operands[0].value = result


@opcode
def SETGE(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Greater than or Equal """
    operands = instruction.operands
    result = int(cpu_context.registers.sf == cpu_context.registers.of)
    logger.debug("Setting %s to 0x%X", operands[0].text, result)
    operands[0].value = result


@opcode
def SETG(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Greather than """
    operands = instruction.operands
    result = int(cpu_context.registers.zf and (cpu_context.registers.sf == cpu_context.registers.of))
    logger.debug("Setting %s to 0x%X", operands[0].text, result)
    operands[0].value = result


@opcode
def SETE(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Equal """
    operands = instruction.operands
    result = int(cpu_context.registers.zf)
    logger.debug("Setting %s to 0x%X", operands[0].text, result)
    operands[0].value = result
    # cpu_context.set_operand_value(0, result)


@opcode
def SETC(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Carry """
    operands = instruction.operands
    result = int(cpu_context.registers.cf)
    logger.debug("Setting %s to 0x%X", operands[0].text, result)
    operands[0].value = result


@opcode
def SETBE(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Below or Equal """
    operands = instruction.operands
    result = int(cpu_context.registers.cf and cpu_context.registers.zf)
    logger.debug("Setting %s to 0x%X", operands[0].text, result)
    operands[0].value = result


@opcode
def SETB(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Below """
    SETC(cpu_context, instruction)


@opcode
def SETAE(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Above or Equal """
    SETC(cpu_context, instruction)


@opcode
def SETA(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Above """
    operands = instruction.operands
    result = int(not (cpu_context.registers.cf | cpu_context.registers.zf))
    logger.debug("Setting %s to 0x%X", operands[0].text, result)
    operands[0].value = result


@opcode
def SETPS(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Not??? Parity """
    operands = instruction.operands
    result = int(cpu_context.registers.sf)
    logger.debug("Setting %s to 0x%X", operands[0].text, result)
    operands[0].value = result


@opcode("setp")
@opcode("setpe")
def setp(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Parity """
    operands = instruction.operands
    result = int(cpu_context.registers.pf)
    logger.debug("Setting %s to 0x%X", operands[0].text, result)
    operands[0].value = result


@opcode("setnp")
@opcode("setpo")
def setnp(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Not Parity """
    operands = instruction.operands
    result = int(not cpu_context.registers.pf)
    logger.debug("Setting %s to 0x%X", operands[0].text, result)
    operands[0].value = result


@opcode
def SETO(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Overflow """
    operands = instruction.operands
    result = int(cpu_context.registers.of)
    logger.debug("Setting %s to 0x%X", operands[0].text, result)
    operands[0].value = result


@opcode
def SETNS(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Not Sign """
    operands = instruction.operands
    result = int(not cpu_context.registers.sf)
    logger.debug("Setting %s to 0x%X", operands[0].text, result)
    operands[0].value = result


@opcode
def SETNO(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Not Overflow """
    operands = instruction.operands
    result = int(not cpu_context.registers.of)
    logger.debug("Setting %s to 0x%X", operands[0].text, result)
    operands[0].value = result


@opcode
def SETNL(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Not Less """
    operands = instruction.operands
    result = int(cpu_context.registers.sf == cpu_context.registers.of)
    logger.debug("Setting %s to 0x%X", operands[0].text, result)
    operands[0].value = result
    # cpu_context.set_operand_value(0, result)


@opcode
def SETNGE(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Not Greater Than or Equal """
    operands = instruction.operands
    result = int(cpu_context.registers.sf != cpu_context.registers.of)
    logger.debug("Setting %s to 0x%X", operands[0].text, result)
    operands[0].value = result


@opcode
def SETNG(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Not Greater """
    operands = instruction.operands
    result = int(cpu_context.registers.zf or (cpu_context.registers.sf != cpu_context.registers.of))
    logger.debug("Setting %s to 0x%X", operands[0].text, result)
    operands[0].value = result


@opcode
def SETNE(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Not Equal """
    operands = instruction.operands
    result = int(not cpu_context.registers.zf)
    logger.debug("Setting %s to 0x%X", operands[0].text, result)
    operands[0].value = result


@opcode
def SETNC(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Not Carry """
    operands = instruction.operands
    result = int(not cpu_context.registers.cf)
    logger.debug("Setting %s to 0x%X", operands[0].text, result)
    operands[0].value = result


@opcode
def SETNBE(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Not Below or Equal """
    operands = instruction.operands
    result = int(not (cpu_context.registers.cf | cpu_context.registers.zf))
    logger.debug("Setting %s to 0x%X", operands[0].text, result)
    operands[0].value = result


@opcode
def SETNB(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Not Below """
    SETNC(cpu_context, instruction)


@opcode
def SETNAE(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Not Above or Equal """
    SETC(cpu_context, instruction)


@opcode
def SETL(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Less Than """
    SETNGE(cpu_context, instruction)


@opcode
def SETNLE(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Not Less Than or Equal """
    SETG(cpu_context, instruction)


@opcode
def SETNZ(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Not Zero """
    SETNE(cpu_context, instruction)


@opcode
def SETZ(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set if Zero """
    SETE(cpu_context, instruction)


@opcode
def SHR(cpu_context: ProcessorContext, instruction: Instruction):
    """ Shift Right """
    operands = instruction.operands
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)

    if not opvalue2:
        return

    result = opvalue1
    tempcount = opvalue2 & (0x3F if cpu_context.bitness == 64 else 0x1F)
    while tempcount:
        cpu_context.registers.cf = get_lsb(result)
        result >>= 1
        tempcount -= 1

    cpu_context.registers.cf = (opvalue1 >> (opvalue2 - 1)) & 0x01
    cpu_context.registers.zf = int(result == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    if opvalue2 == 1:
        cpu_context.registers.of = utils.sign_bit(opvalue1, width)
    else:
        cpu_context.registers.of = 0

    cpu_context.registers.pf = get_parity(result)
    if cpu_context.emulator.branch_tracking:
        cpu_context.jcccontext.update_flag_opnds(["cf", "zf", "sf", "of", "pf"], operands)
    logger.debug("Shift 0x%X right by 0x%X -> 0x%X", opvalue1, opvalue2, result)
    operands[0].value = result


@opcode
def STD(cpu_context: ProcessorContext, instruction: Instruction):
    """ Set Direction Flag """
    cpu_context.registers.df = 1


@opcode
def STOSB(cpu_context: ProcessorContext, instruction: Instruction):
    """Store value in AL in the address pointed to by EDI"""
    value = cpu_context.registers.al
    addr = cpu_context.registers.edi
    logger.debug("Storing 0x%X into 0x%X", value, addr)
    cpu_context.memory.write(addr, bytes([value]))
    if cpu_context.registers.df:
        cpu_context.registers.edi -= 1
    else:
        cpu_context.registers.edi += 1


@opcode
def STOSW(cpu_context: ProcessorContext, instruction: Instruction):
    """Store value in AX in the address pointed to by EDI"""
    value = cpu_context.registers.ax
    addr = cpu_context.registers.edi
    logger.debug("Storing 0x%X into 0x%X", value, addr)
    cpu_context.memory.write(addr, value.to_bytes(2, cpu_context.byteorder))
    if cpu_context.registers.df:
        cpu_context.registers.edi -= 2
    else:
        cpu_context.registers.edi += 2


@opcode
def STOSD(cpu_context: ProcessorContext, instruction: Instruction):
    """Store value in EAX in the address pointed to by EDI"""
    value = cpu_context.registers.eax
    addr = cpu_context.registers.edi
    logger.debug("Storing 0x%X into 0x%X", value, addr)
    cpu_context.memory.write(addr, value.to_bytes(4, cpu_context.byteorder))
    if cpu_context.registers.df:
        cpu_context.registers.edi -= 4
    else:
        cpu_context.registers.edi += 4


@opcode
def STOSQ(cpu_context: ProcessorContext, instruction: Instruction):
    """Store value in RAX in the address pointed to by RDI"""
    value = cpu_context.registers.rax
    addr = cpu_context.registers.rdi
    logger.debug("Storing 0x%X into 0x%X", value, addr)
    cpu_context.memory.write(addr, value.to_bytes(8, cpu_context.byteorder))
    if cpu_context.registers.df:
        cpu_context.registers.rdi -= 8
    else:
        cpu_context.registers.rdi += 8


@opcode
def SUB(cpu_context: ProcessorContext, instruction: Instruction):
    """ Subtract """
    operands = instruction.operands
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)
    result = opvalue1 - opvalue2

    mask = utils.get_mask(width)
    cpu_context.registers.cf = int((opvalue1 & mask) < (opvalue2 & mask))
    cpu_context.registers.af = int((opvalue1 ^ opvalue2 ^ result) & 0x10)
    cpu_context.registers.zf = int(result & mask == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = int(not (-(mask // 2) <= result < (mask // 2)))
    cpu_context.registers.pf = get_parity(result)
    if cpu_context.emulator.branch_tracking:
        cpu_context.jcccontext.update_flag_opnds(["cf", "af", "zf", "sf", "of", "pf"], operands)

    logger.debug("0x%X - 0x%X = 0x%X", opvalue1, opvalue2, result)
    operands[0].value = result


@opcode
def TEST(cpu_context: ProcessorContext, instruction: Instruction):
    """ Test values for equality """
    operands = instruction.operands
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)
    result = opvalue1 & opvalue2

    mask = utils.get_mask(width)
    cpu_context.registers.cf = 0
    cpu_context.registers.af = int((opvalue1 ^ opvalue2 ^ result) & 0x10)
    cpu_context.registers.zf = int(result & mask == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = 0
    cpu_context.registers.pf = get_parity(result)
    if cpu_context.emulator.branch_tracking:
        cpu_context.jcccontext.update_flag_opnds(["cf", "af", "zf", "sf", "of", "pf"], operands)

    logger.debug("0x%X & 0x%X -> 0x%X", opvalue1, opvalue2, result)


@opcode
def XCHG(cpu_context: ProcessorContext, instruction: Instruction):
    """ Exchange two values """
    operands = instruction.operands
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    logger.debug("exchange 0x%X and 0x%X", opvalue1, opvalue2)
    operands[1].value = opvalue1
    operands[0].value = opvalue2


@opcode("xor")
@opcode("pxor")
def _xor(cpu_context: ProcessorContext, instruction: Instruction):
    """ XOR """
    operands = instruction.operands
    opvalue1 = operands[0].value
    opvalue2 = operands[1].value
    width = get_max_operand_size(operands)
    result = opvalue1 ^ opvalue2

    cpu_context.registers.cf = 0
    cpu_context.registers.zf = int(result == 0)
    cpu_context.registers.sf = utils.sign_bit(result, width)
    cpu_context.registers.of = 0
    cpu_context.registers.pf = get_parity(result)
    if cpu_context.emulator.branch_tracking:
        cpu_context.jcccontext.update_flag_opnds(["cf", "zf", "sf", "of", "pf"], operands)

    logger.debug("0x%X ^ 0x%X = 0x%X", opvalue1, opvalue2, result)
    operands[0].value = result


# Global helper functions


def get_max_operand_size(operands):
    """
    Given the list of named tuples containing the operand value and bit width, determine the largest bit width.

    :param operands: list of Operand objects

    :return: largest operand width
    """
    return max(operand.width for operand in operands)


def get_min_operand_size(operands):
    """
    Given the list of named tuples containing the operand value and bit width, determine the smallest bit width.

    :param operands: list of Operand objects

    :return: smallest operand width
    """
    return min(operand.width for operand in operands)


def get_msb(value, size):
    """
    Get most significant bit.

    :param value: value to obtain msb from
    :param size: bit width of value in bytes

    :return: most significant bit
    :raises AssertionError: value is larger than given size.
    """
    msb = value >> ((8 * size) - 1)
    if msb > 1:
        raise AssertionError(f"Got invalid size {size} for value 0x{value:x}")
    return msb


def get_lsb(value):
    """
    Get least significant bit.

    :param value: value to obtain lsb from

    :return: least significant bit
    """
    return value & 0x1


def swap_bytes(value, size):
    """
    Swaps a set of bytes based on size

    :param value: value to swap

    :param size: width of value in bytes

    :return: swapped
    """
    if size == 1:
        return value

    if size == 2:
        return ((value & 0xFF) << 8) | ((value & 0xFF00) >> 8)

    if size == 4:
        return (
            ((value & 0xFF) << 24)
            | (((value & 0xFF00) >> 8) << 16)
            | (((value & 0xFF0000) >> 16) << 8)
            | ((value & 0xFF000000) >> 24)
        )

    if size == 8:
        return (
            ((value & 0xFF) << 56)
            | (((value & 0xFF00) >> 8) << 48)
            | (((value & 0xFF0000) >> 16) << 40)
            | (((value & 0xFF000000) >> 24) << 32)
            | (((value & 0xFF00000000) >> 32) << 24)
            | (((value & 0xFF0000000000) >> 40) << 16)
            | (((value & 0xFF000000000000) >> 48) << 8)
            | ((value & 0xFF00000000000000) >> 56)
        )


# fmt: off
parity_lookup_table = [
    1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
    0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
    0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
    1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
    0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
    1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
    1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
    0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
    0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
    1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
    1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
    0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
    1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
    0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
    0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
    1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
]
# fmt: on


def get_parity(value):
    """Returns the parity of the given value."""
    return parity_lookup_table[value & 0xFF]



"""
Interface for function management.
"""
from __future__ import annotations

from typing import List

import logging

import dragodis.interface
from dragodis import Disassembler, NotExistError
from dragodis.interface import (
    StackLocation, RelativeRegisterLocation, StaticLocation, RegisterLocation,
    RegisterPairLocation, ArgumentLocation
)

from . import utils

logger = logging.getLogger(__name__)


class FunctionSignature:
    """
    Interface for a function signature.

    NOTE: This object retrieves the signature on initialization.
    Any external changes to the function's signature after this object is created will take
    no affect.
    As well, any changes done to this object will not affect the signature in the IDB unless
    the apply() function is called.
    """

    def __init__(self, cpu_context, address: int, signature: dragodis.interface.FunctionSignature):
        """
        :param cpu_context: ProcessorContext to use for pulling argument values.
        :param address: Address of function for this signature.
        :param signature: dragodis FunctionSignature to wrap.
        """
        self._cpu_context = cpu_context
        self._signature = signature
        self.address = address

    def __repr__(self):
        return f'< FunctionSignature : {self.declaration} >'

    @property
    def name(self) -> str:
        """The demangled name of function."""
        return self._signature.name

    @property
    def declaration(self) -> str:
        """The full function declaration."""
        return self._signature.declaration

    @property
    def calling_convention(self) -> str:
        """
        The calling convention used in the function signature.
        """
        return self._signature.calling_convention

    @calling_convention.setter
    def calling_convention(self, name: str):
        """
        Sets the calling convention for the function signature.
        """
        self._signature.calling_convention = name

    @property
    def return_type(self) -> str:
        """
        The return type of the function.
        """
        return self._signature.return_type.name

    @return_type.setter
    def return_type(self, data_type: str):
        """
        Sets the return type of the function with given data type.
        """
        self._signature.return_type = data_type

    @property
    def arguments(self) -> List[FunctionArgument]:
        """
        Gets the defined arguments for the function signature.
        NOTE: We name this argument instead of parameters since it is based on an
        underlying context for a specific value.
        """
        return [
            FunctionArgument(self._cpu_context, self, parameter)
            for parameter in self._signature.parameters
        ]

    def remove_argument(self, ordinal: int):
        self._signature.remove_parameter(ordinal)

    def add_argument(self, data_type: str):
        self._signature.add_parameter(data_type)

    def insert_argument(self, ordinal: int, data_type: str):
        self._signature.insert_parameter(ordinal, data_type)

    def replace_arguments(self, data_types: List[str]):
        self._signature.replace_parameters(data_types)


class FunctionArgument:
    """
    Interface for a function argument from FunctionSignature
    """

    def __init__(self, cpu_context, signature: FunctionSignature, parameter: dragodis.interface.FunctionParameter):
        self._cpu_context = cpu_context
        self._signature = signature
        self._parameter = parameter

    def __repr__(self):
        string = (
            f"< FunctionArg {hex(self._signature.address)}:{self.ordinal} "
            f": {self.declaration} = {hex(self.value)}"
        )
        if self.addr is not None:
            string += f" : &{self.name} = {hex(self.addr)}"
        string += " >"
        return string

    @property
    def ordinal(self) -> int:
        return self._parameter.ordinal

    @property
    def width(self) -> int:
        return self._parameter.size

    @property
    def name(self) -> str:
        return self._parameter.name

    @name.setter
    def name(self, value: str):
        self._parameter.name = value

    @property
    def location(self) -> ArgumentLocation:
        return self._parameter.location

    @property
    def type(self) -> str:
        """User friendly type name."""
        return self._parameter.data_type.name

    @type.setter
    def type(self, value: str):
        """
        Sets function argument to a new type.
        """
        # TODO: Should modifying the data type affect the signature itself?
        self._parameter.data_type = value

    @property
    def declaration(self):
        """Argument type declaration."""
        return self._parameter.declaration

    @property
    def is_stack(self):
        """True if argument is on the stack."""
        return isinstance(self.location, StackLocation)

    # TODO: Refactor to be more processor agnostic
    @property
    def addr(self):
        """Retrieves the address of the argument (if a memory/stack address)"""
        location = self.location
        disassembler: Disassembler = self._cpu_context.emulator.disassembler

        if isinstance(location, StackLocation):
            # First get the offsetted stack location.
            addr = self._cpu_context.sp + location.stack_offset

            try:
                func = disassembler.get_function(self._signature.address)
                in_function = self._cpu_context.ip in func
            except NotExistError:
                in_function = False

            if disassembler.processor_name == "x86":
                # If we are inside the calling function (or in the middle of hooking a call)
                # account for the pushed in return address.
                if in_function or self._cpu_context.hooking_call == self._signature.address:
                    # Determine adjustment of stack based on what IDA reports as the current
                    # ESP plus the size of the saved return address.
                    # (More reliable and cross compatible than using ebp.)
                    retaddr_size = self._cpu_context.byteness
                    addr += retaddr_size

            # Also adjust the stack if we are in the function.
            if in_function:
                instruction = disassembler.get_instruction(self._cpu_context.ip)
                addr -= instruction.stack_depth  # this is negative

            return addr

        if isinstance(location, StaticLocation):
            return location.address

        return None

    @property
    def value(self):
        """Retrieves the value of the argument based on the cpu context."""
        # TODO: Pull value data based on type.
        location = self.location

        # On Stack
        if isinstance(location, StackLocation):
            logger.debug(
                f"Retrieving argument {self.ordinal} at {hex(self.addr)}, "
                f"stack: {hex(self._cpu_context.sp)}, offset: {hex(location.stack_offset)}"
            )
            # read the argument from teh stack using the calculated stack offset from the disassembler.
            value = self._cpu_context.memory.read(self.addr, self._cpu_context.byteness)
            return int.from_bytes(value, self._cpu_context.byteorder)

        # Single register
        if isinstance(location, RegisterLocation):
            return self._cpu_context.registers[location.register.name]

        # Register pair (eg: edx:eax)
        if isinstance(location, RegisterPairLocation):
            # Width is the combination of both registers.
            reg_width = self.width // 2
            reg1, reg2 = location.registers
            logger.debug("Register pair: [%s:%s]", reg1.name, reg2.name)
            reg1_value = self._cpu_context.registers[reg1.name]
            reg2_value = self._cpu_context.registers[reg2.name]
            return (reg2_value << (reg_width * 8)) | reg1_value

        # Relative register (displacement from address pointed by register)
        if isinstance(location, RelativeRegisterLocation):
            # TODO: CURRENTLY UNTESTED
            logger.info(
                f"Argument {self.ordinal} of untested type ALOC_RREL.  "
                f"Verify results and report issues."
            )
            return self._cpu_context.registers[location.register.name] + location.offset

        # Global address
        if isinstance(location, StaticLocation):
            return location.address

        raise NotImplementedError(f"Unsupported location: {location!r}")

    @value.setter
    def value(self, value):
        """Sets the value of the argument to the cpu context."""
        # TODO: Pull value data based on type.
        location = self.location

        # On Stack
        if isinstance(location, StackLocation):
            logger.debug(
                f"Setting argument {self.ordinal} at {hex(self.addr)}, "
                f"stack: {hex(self._cpu_context.sp)}, offset: {hex(location.stack_offset)}"
            )
            data = value.to_bytes(self._cpu_context.byteness, self._cpu_context.byteorder)
            self._cpu_context.memory.write(self.addr, data)

        # Single register
        elif isinstance(location, RegisterLocation):
            self._cpu_context.registers[location.register.name] = value

        # Register pair (eg: edx:eax)
        elif isinstance(location, RegisterPairLocation):
            reg_width = self.width // 2
            reg1, reg2 = location.registers
            self._cpu_context.registers[reg1.name] = value & utils.get_mask(reg_width)
            self._cpu_context.registers[reg2.name] = value >> (reg_width * 8)

        # Relative register (displacement from address pointed by register)
        elif isinstance(location, RelativeRegisterLocation):
            # TODO: CURRENTLY UNTESTED
            logger.info(
                f"Argument {self.ordinal} of untested type ALOC_RREL.  "
                r"Verify results and report issues."
            )
            self._cpu_context.registers[location.register.name] = value - location.offset

        # Global address
        elif isinstance(location, StaticLocation):
            raise TypeError(f"Unable to set a argument {self.ordinal} with a static location.")

        else:
            raise NotImplementedError(f"Unsupported location: {location!r}")



"""
Helper utilities for functions.
"""
import collections
import warnings

import dragodis
from dragodis.interface import Function

__all__ = [
    "from_name",
    "api_calls",
]


def from_name(dis: dragodis.Disassembler, func_name: str, ignore_underscore: bool = False) -> Function:
    warnings.warn(
        "func_utils.from_name() is deprecated. Please use Disassembler.get_function_by_name() instead.",
        DeprecationWarning
    )
    try:
        return dis.get_function_by_name(func_name, ignore_underscore=ignore_underscore)
    except dragodis.NotExistError as e:
        raise ValueError(str(e))


@property
def api_calls(function: Function) -> collections.Counter:
    """
    Returns counter containing API calls and the number of times they were called.

    e.g.
        with dragodis.open_program("input.exe") as dis:
            func = dis.get_function(0x40000)
            api_calls = functions.api_calls(func)

    """
    return collections.Counter(callee.name for callee in function.callees)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'emulator.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_EmulatorForm(object):
    def setupUi(self, EmulatorForm):
        EmulatorForm.setObjectName("EmulatorForm")
        EmulatorForm.resize(770, 796)
        self.verticalLayout = QtWidgets.QVBoxLayout(EmulatorForm)
        self.verticalLayout.setObjectName("verticalLayout")
        self.emulator_controls = QtWidgets.QHBoxLayout()
        self.emulator_controls.setObjectName("emulator_controls")
        self.trace_depth_label = QtWidgets.QLabel(EmulatorForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.trace_depth_label.sizePolicy().hasHeightForWidth())
        self.trace_depth_label.setSizePolicy(sizePolicy)
        self.trace_depth_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.trace_depth_label.setObjectName("trace_depth_label")
        self.emulator_controls.addWidget(self.trace_depth_label)
        self.trace_depth = QtWidgets.QSpinBox(EmulatorForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.trace_depth.sizePolicy().hasHeightForWidth())
        self.trace_depth.setSizePolicy(sizePolicy)
        self.trace_depth.setObjectName("trace_depth")
        self.emulator_controls.addWidget(self.trace_depth)
        self.call_depth_label = QtWidgets.QLabel(EmulatorForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.call_depth_label.sizePolicy().hasHeightForWidth())
        self.call_depth_label.setSizePolicy(sizePolicy)
        self.call_depth_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.call_depth_label.setObjectName("call_depth_label")
        self.emulator_controls.addWidget(self.call_depth_label)
        self.call_depth = QtWidgets.QSpinBox(EmulatorForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.call_depth.sizePolicy().hasHeightForWidth())
        self.call_depth.setSizePolicy(sizePolicy)
        self.call_depth.setObjectName("call_depth")
        self.emulator_controls.addWidget(self.call_depth)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.emulator_controls.addItem(spacerItem)
        self.exhaustive = QtWidgets.QCheckBox(EmulatorForm)
        self.exhaustive.setChecked(True)
        self.exhaustive.setObjectName("exhaustive")
        self.emulator_controls.addWidget(self.exhaustive)
        self.follow_loops = QtWidgets.QCheckBox(EmulatorForm)
        self.follow_loops.setObjectName("follow_loops")
        self.emulator_controls.addWidget(self.follow_loops)
        self.verticalLayout.addLayout(self.emulator_controls)
        self.emulator_buttons = QtWidgets.QHBoxLayout()
        self.emulator_buttons.setObjectName("emulator_buttons")
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.emulator_buttons.addItem(spacerItem1)
        self.run_button = QtWidgets.QPushButton(EmulatorForm)
        self.run_button.setObjectName("run_button")
        self.emulator_buttons.addWidget(self.run_button)
        self.step_over_button = QtWidgets.QPushButton(EmulatorForm)
        self.step_over_button.setEnabled(False)
        self.step_over_button.setObjectName("step_over_button")
        self.emulator_buttons.addWidget(self.step_over_button)
        self.step_into_button = QtWidgets.QPushButton(EmulatorForm)
        self.step_into_button.setEnabled(False)
        self.step_into_button.setObjectName("step_into_button")
        self.emulator_buttons.addWidget(self.step_into_button)
        self.step_out_button = QtWidgets.QPushButton(EmulatorForm)
        self.step_out_button.setEnabled(False)
        self.step_out_button.setObjectName("step_out_button")
        self.emulator_buttons.addWidget(self.step_out_button)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.emulator_buttons.addItem(spacerItem2)
        self.verticalLayout.addLayout(self.emulator_buttons)
        self.instruction = QtWidgets.QLabel(EmulatorForm)
        self.instruction.setObjectName("instruction")
        self.verticalLayout.addWidget(self.instruction)
        self.tabs = QtWidgets.QTabWidget(EmulatorForm)
        self.tabs.setTabPosition(QtWidgets.QTabWidget.North)
        self.tabs.setTabShape(QtWidgets.QTabWidget.Rounded)
        self.tabs.setDocumentMode(True)
        self.tabs.setTabsClosable(False)
        self.tabs.setMovable(True)
        self.tabs.setObjectName("tabs")
        self.operands_tab = QtWidgets.QWidget()
        self.operands_tab.setObjectName("operands_tab")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.operands_tab)
        self.verticalLayout_2.setContentsMargins(0, 6, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.operands_table = QtWidgets.QTableWidget(self.operands_tab)
        self.operands_table.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.operands_table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.operands_table.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.operands_table.setObjectName("operands_table")
        self.operands_table.setColumnCount(6)
        self.operands_table.setRowCount(0)
        item = QtWidgets.QTableWidgetItem()
        self.operands_table.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.operands_table.setHorizontalHeaderItem(1, item)
        item = QtWidgets.QTableWidgetItem()
        self.operands_table.setHorizontalHeaderItem(2, item)
        item = QtWidgets.QTableWidgetItem()
        self.operands_table.setHorizontalHeaderItem(3, item)
        item = QtWidgets.QTableWidgetItem()
        self.operands_table.setHorizontalHeaderItem(4, item)
        item = QtWidgets.QTableWidgetItem()
        self.operands_table.setHorizontalHeaderItem(5, item)
        self.operands_table.horizontalHeader().setCascadingSectionResizes(True)
        self.operands_table.verticalHeader().setVisible(False)
        self.verticalLayout_2.addWidget(self.operands_table)
        self.tabs.addTab(self.operands_tab, "")
        self.registers_tab = QtWidgets.QWidget()
        self.registers_tab.setObjectName("registers_tab")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.registers_tab)
        self.verticalLayout_3.setContentsMargins(0, 6, 0, 0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.registers_table = QtWidgets.QTableWidget(self.registers_tab)
        self.registers_table.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.registers_table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.registers_table.setObjectName("registers_table")
        self.registers_table.setColumnCount(3)
        self.registers_table.setRowCount(0)
        item = QtWidgets.QTableWidgetItem()
        self.registers_table.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.registers_table.setHorizontalHeaderItem(1, item)
        item = QtWidgets.QTableWidgetItem()
        self.registers_table.setHorizontalHeaderItem(2, item)
        self.registers_table.verticalHeader().setVisible(False)
        self.verticalLayout_3.addWidget(self.registers_table)
        self.tabs.addTab(self.registers_tab, "")
        self.memory_tab = QtWidgets.QWidget()
        self.memory_tab.setObjectName("memory_tab")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.memory_tab)
        self.verticalLayout_6.setContentsMargins(0, 6, 0, 0)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.memory_controls = QtWidgets.QHBoxLayout()
        self.memory_controls.setContentsMargins(-1, 3, -1, 0)
        self.memory_controls.setObjectName("memory_controls")
        self.label = QtWidgets.QLabel(self.memory_tab)
        self.label.setObjectName("label")
        self.memory_controls.addWidget(self.label)
        self.memory_start = QtWidgets.QLineEdit(self.memory_tab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.memory_start.sizePolicy().hasHeightForWidth())
        self.memory_start.setSizePolicy(sizePolicy)
        self.memory_start.setObjectName("memory_start")
        self.memory_controls.addWidget(self.memory_start)
        self.label_2 = QtWidgets.QLabel(self.memory_tab)
        self.label_2.setObjectName("label_2")
        self.memory_controls.addWidget(self.label_2)
        self.memory_size = QtWidgets.QSpinBox(self.memory_tab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.memory_size.sizePolicy().hasHeightForWidth())
        self.memory_size.setSizePolicy(sizePolicy)
        self.memory_size.setMinimumSize(QtCore.QSize(10, 0))
        self.memory_size.setMaximum(65535)
        self.memory_size.setProperty("value", 1024)
        self.memory_size.setObjectName("memory_size")
        self.memory_controls.addWidget(self.memory_size)
        self.memory_load_button = QtWidgets.QPushButton(self.memory_tab)
        self.memory_load_button.setObjectName("memory_load_button")
        self.memory_controls.addWidget(self.memory_load_button)
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.memory_controls.addItem(spacerItem3)
        self.memory_export_button = QtWidgets.QPushButton(self.memory_tab)
        self.memory_export_button.setObjectName("memory_export_button")
        self.memory_controls.addWidget(self.memory_export_button)
        self.verticalLayout_6.addLayout(self.memory_controls)
        self.memory_hexdump = QtWidgets.QTextEdit(self.memory_tab)
        self.memory_hexdump.setLineWrapMode(QtWidgets.QTextEdit.NoWrap)
        self.memory_hexdump.setReadOnly(True)
        self.memory_hexdump.setAcceptRichText(False)
        self.memory_hexdump.setTextInteractionFlags(QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.memory_hexdump.setObjectName("memory_hexdump")
        self.verticalLayout_6.addWidget(self.memory_hexdump)
        self.memory_blocks_table = QtWidgets.QTableWidget(self.memory_tab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.memory_blocks_table.sizePolicy().hasHeightForWidth())
        self.memory_blocks_table.setSizePolicy(sizePolicy)
        self.memory_blocks_table.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
        self.memory_blocks_table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.memory_blocks_table.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.memory_blocks_table.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.memory_blocks_table.setObjectName("memory_blocks_table")
        self.memory_blocks_table.setColumnCount(3)
        self.memory_blocks_table.setRowCount(0)
        item = QtWidgets.QTableWidgetItem()
        self.memory_blocks_table.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.memory_blocks_table.setHorizontalHeaderItem(1, item)
        item = QtWidgets.QTableWidgetItem()
        self.memory_blocks_table.setHorizontalHeaderItem(2, item)
        self.memory_blocks_table.verticalHeader().setVisible(False)
        self.verticalLayout_6.addWidget(self.memory_blocks_table)
        self.tabs.addTab(self.memory_tab, "")
        self.variables_tab = QtWidgets.QWidget()
        self.variables_tab.setObjectName("variables_tab")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.variables_tab)
        self.verticalLayout_4.setContentsMargins(0, 6, 0, 0)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.variables_table = QtWidgets.QTableWidget(self.variables_tab)
        self.variables_table.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.variables_table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.variables_table.setObjectName("variables_table")
        self.variables_table.setColumnCount(6)
        self.variables_table.setRowCount(0)
        item = QtWidgets.QTableWidgetItem()
        self.variables_table.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.variables_table.setHorizontalHeaderItem(1, item)
        item = QtWidgets.QTableWidgetItem()
        self.variables_table.setHorizontalHeaderItem(2, item)
        item = QtWidgets.QTableWidgetItem()
        self.variables_table.setHorizontalHeaderItem(3, item)
        item = QtWidgets.QTableWidgetItem()
        self.variables_table.setHorizontalHeaderItem(4, item)
        item = QtWidgets.QTableWidgetItem()
        self.variables_table.setHorizontalHeaderItem(5, item)
        self.variables_table.horizontalHeader().setCascadingSectionResizes(True)
        self.variables_table.horizontalHeader().setSortIndicatorShown(True)
        self.variables_table.verticalHeader().setVisible(False)
        self.variables_table.verticalHeader().setCascadingSectionResizes(False)
        self.verticalLayout_4.addWidget(self.variables_table)
        self.tabs.addTab(self.variables_tab, "")
        self.function_arguments_tab = QtWidgets.QWidget()
        self.function_arguments_tab.setObjectName("function_arguments_tab")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.function_arguments_tab)
        self.verticalLayout_5.setContentsMargins(0, 6, 0, 0)
        self.verticalLayout_5.setSpacing(6)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.function_arguments_form = QtWidgets.QHBoxLayout()
        self.function_arguments_form.setContentsMargins(3, 3, 3, 3)
        self.function_arguments_form.setObjectName("function_arguments_form")
        self.function_signature = QtWidgets.QLabel(self.function_arguments_tab)
        self.function_signature.setObjectName("function_signature")
        self.function_arguments_form.addWidget(self.function_signature)
        spacerItem4 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.function_arguments_form.addItem(spacerItem4)
        self.num_args_label = QtWidgets.QLabel(self.function_arguments_tab)
        self.num_args_label.setObjectName("num_args_label")
        self.function_arguments_form.addWidget(self.num_args_label)
        self.num_args = QtWidgets.QSpinBox(self.function_arguments_tab)
        self.num_args.setObjectName("num_args")
        self.function_arguments_form.addWidget(self.num_args)
        self.verticalLayout_5.addLayout(self.function_arguments_form)
        self.function_arguments_table = QtWidgets.QTableWidget(self.function_arguments_tab)
        self.function_arguments_table.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.function_arguments_table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.function_arguments_table.setObjectName("function_arguments_table")
        self.function_arguments_table.setColumnCount(8)
        self.function_arguments_table.setRowCount(0)
        item = QtWidgets.QTableWidgetItem()
        self.function_arguments_table.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.function_arguments_table.setHorizontalHeaderItem(1, item)
        item = QtWidgets.QTableWidgetItem()
        self.function_arguments_table.setHorizontalHeaderItem(2, item)
        item = QtWidgets.QTableWidgetItem()
        self.function_arguments_table.setHorizontalHeaderItem(3, item)
        item = QtWidgets.QTableWidgetItem()
        self.function_arguments_table.setHorizontalHeaderItem(4, item)
        item = QtWidgets.QTableWidgetItem()
        self.function_arguments_table.setHorizontalHeaderItem(5, item)
        item = QtWidgets.QTableWidgetItem()
        self.function_arguments_table.setHorizontalHeaderItem(6, item)
        item = QtWidgets.QTableWidgetItem()
        self.function_arguments_table.setHorizontalHeaderItem(7, item)
        self.function_arguments_table.verticalHeader().setVisible(False)
        self.verticalLayout_5.addWidget(self.function_arguments_table)
        self.tabs.addTab(self.function_arguments_tab, "")
        self.call_history_tab = QtWidgets.QWidget()
        self.call_history_tab.setObjectName("call_history_tab")
        self.verticalLayout_8 = QtWidgets.QVBoxLayout(self.call_history_tab)
        self.verticalLayout_8.setContentsMargins(0, 6, 0, 0)
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.call_history_table = QtWidgets.QTableWidget(self.call_history_tab)
        self.call_history_table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.call_history_table.setObjectName("call_history_table")
        self.call_history_table.setColumnCount(2)
        self.call_history_table.setRowCount(0)
        item = QtWidgets.QTableWidgetItem()
        self.call_history_table.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.call_history_table.setHorizontalHeaderItem(1, item)
        self.call_history_table.verticalHeader().setVisible(False)
        self.verticalLayout_8.addWidget(self.call_history_table)
        self.tabs.addTab(self.call_history_tab, "")
        self.actions_tab = QtWidgets.QWidget()
        self.actions_tab.setObjectName("actions_tab")
        self.actions_layout = QtWidgets.QVBoxLayout(self.actions_tab)
        self.actions_layout.setContentsMargins(0, 6, 0, 0)
        self.actions_layout.setObjectName("actions_layout")
        self.actions_table = QtWidgets.QTreeWidget(self.actions_tab)
        self.actions_table.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.actions_table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.actions_table.setIndentation(20)
        self.actions_table.setRootIsDecorated(True)
        self.actions_table.setUniformRowHeights(False)
        self.actions_table.setAnimated(True)
        self.actions_table.setHeaderHidden(False)
        self.actions_table.setObjectName("actions_table")
        self.actions_table.header().setSortIndicatorShown(True)
        self.actions_layout.addWidget(self.actions_table)
        self.tabs.addTab(self.actions_tab, "")
        self.stdout_tab = QtWidgets.QWidget()
        self.stdout_tab.setObjectName("stdout_tab")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.stdout_tab)
        self.verticalLayout_7.setContentsMargins(0, 6, 0, 0)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem5 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem5)
        self.stdout_export_button = QtWidgets.QPushButton(self.stdout_tab)
        self.stdout_export_button.setObjectName("stdout_export_button")
        self.horizontalLayout.addWidget(self.stdout_export_button)
        self.verticalLayout_7.addLayout(self.horizontalLayout)
        self.stdout_textdump = QtWidgets.QTextEdit(self.stdout_tab)
        self.stdout_textdump.setLineWrapMode(QtWidgets.QTextEdit.NoWrap)
        self.stdout_textdump.setReadOnly(True)
        self.stdout_textdump.setAcceptRichText(False)
        self.stdout_textdump.setTextInteractionFlags(QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.stdout_textdump.setObjectName("stdout_textdump")
        self.verticalLayout_7.addWidget(self.stdout_textdump)
        self.tabs.addTab(self.stdout_tab, "")
        self.verticalLayout.addWidget(self.tabs)
        self.status = QtWidgets.QLabel(EmulatorForm)
        self.status.setObjectName("status")
        self.verticalLayout.addWidget(self.status)

        self.retranslateUi(EmulatorForm)
        self.tabs.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(EmulatorForm)

    def retranslateUi(self, EmulatorForm):
        _translate = QtCore.QCoreApplication.translate
        EmulatorForm.setWindowTitle(_translate("EmulatorForm", "Rugosa Emulator"))
        self.trace_depth_label.setToolTip(_translate("EmulatorForm", "Number of levels of parent callers to emulate"))
        self.trace_depth_label.setText(_translate("EmulatorForm", "Trace Depth"))
        self.call_depth_label.setToolTip(_translate("EmulatorForm", "Number of call levels deep to emulate"))
        self.call_depth_label.setText(_translate("EmulatorForm", "Call Depth"))
        self.exhaustive.setToolTip(_translate("EmulatorForm", "Whether to also follow loops for parent callers"))
        self.exhaustive.setText(_translate("EmulatorForm", "Exhaustive"))
        self.follow_loops.setToolTip(_translate("EmulatorForm", "Whether to emulate loops instead of forcing a downward path"))
        self.follow_loops.setText(_translate("EmulatorForm", "Follow Loops"))
        self.run_button.setToolTip(_translate("EmulatorForm", "Emulate up to the current highlighted instruction"))
        self.run_button.setText(_translate("EmulatorForm", "Run"))
        self.step_over_button.setToolTip(_translate("EmulatorForm", "Emulate the next instruction"))
        self.step_over_button.setText(_translate("EmulatorForm", "Step Over"))
        self.step_into_button.setToolTip(_translate("EmulatorForm", "Not Supported Yet"))
        self.step_into_button.setText(_translate("EmulatorForm", "Step Into"))
        self.step_out_button.setToolTip(_translate("EmulatorForm", "Not Supported Yet"))
        self.step_out_button.setText(_translate("EmulatorForm", "Step Out"))
        self.instruction.setToolTip(_translate("EmulatorForm", "<html><head/><body><p>Current instruction emulated up to.</p></body></html>"))
        self.instruction.setText(_translate("EmulatorForm", "<instruction>"))
        self.operands_tab.setToolTip(_translate("EmulatorForm", "Contents of the operands of the current instruction"))
        self.operands_table.setSortingEnabled(True)
        item = self.operands_table.horizontalHeaderItem(0)
        item.setText(_translate("EmulatorForm", "Index"))
        item = self.operands_table.horizontalHeaderItem(1)
        item.setText(_translate("EmulatorForm", "Size"))
        item = self.operands_table.horizontalHeaderItem(2)
        item.setText(_translate("EmulatorForm", "Text"))
        item = self.operands_table.horizontalHeaderItem(3)
        item.setText(_translate("EmulatorForm", "Address"))
        item = self.operands_table.horizontalHeaderItem(4)
        item.setText(_translate("EmulatorForm", "Value"))
        item = self.operands_table.horizontalHeaderItem(5)
        item.setText(_translate("EmulatorForm", "Referenced Data"))
        self.tabs.setTabText(self.tabs.indexOf(self.operands_tab), _translate("EmulatorForm", "Operands"))
        self.registers_tab.setToolTip(_translate("EmulatorForm", "Content of the registers"))
        self.registers_table.setSortingEnabled(True)
        item = self.registers_table.horizontalHeaderItem(0)
        item.setText(_translate("EmulatorForm", "Name"))
        item = self.registers_table.horizontalHeaderItem(1)
        item.setText(_translate("EmulatorForm", "Value"))
        item = self.registers_table.horizontalHeaderItem(2)
        item.setText(_translate("EmulatorForm", "Referenced Data"))
        self.tabs.setTabText(self.tabs.indexOf(self.registers_tab), _translate("EmulatorForm", "Registers"))
        self.memory_tab.setToolTip(_translate("EmulatorForm", "Contents of memory"))
        self.label.setText(_translate("EmulatorForm", "Start Address"))
        self.memory_start.setPlaceholderText(_translate("EmulatorForm", "0x401000"))
        self.label_2.setText(_translate("EmulatorForm", "Size"))
        self.memory_load_button.setText(_translate("EmulatorForm", "Load"))
        self.memory_export_button.setText(_translate("EmulatorForm", "Export"))
        self.memory_blocks_table.setToolTip(_translate("EmulatorForm", "Currently mapped memory blocks"))
        item = self.memory_blocks_table.horizontalHeaderItem(0)
        item.setText(_translate("EmulatorForm", "Start"))
        item = self.memory_blocks_table.horizontalHeaderItem(1)
        item.setText(_translate("EmulatorForm", "End"))
        item = self.memory_blocks_table.horizontalHeaderItem(2)
        item.setText(_translate("EmulatorForm", "Size"))
        self.tabs.setTabText(self.tabs.indexOf(self.memory_tab), _translate("EmulatorForm", "Memory"))
        self.variables_tab.setToolTip(_translate("EmulatorForm", "Contents of labeled data"))
        self.variables_table.setSortingEnabled(True)
        item = self.variables_table.horizontalHeaderItem(0)
        item.setText(_translate("EmulatorForm", "Address"))
        item = self.variables_table.horizontalHeaderItem(1)
        item.setText(_translate("EmulatorForm", "Stack Offset"))
        item = self.variables_table.horizontalHeaderItem(2)
        item.setText(_translate("EmulatorForm", "Data Type"))
        item = self.variables_table.horizontalHeaderItem(3)
        item.setText(_translate("EmulatorForm", "Size"))
        item = self.variables_table.horizontalHeaderItem(4)
        item.setText(_translate("EmulatorForm", "Name"))
        item = self.variables_table.horizontalHeaderItem(5)
        item.setText(_translate("EmulatorForm", "Value"))
        self.tabs.setTabText(self.tabs.indexOf(self.variables_tab), _translate("EmulatorForm", "Variables"))
        self.function_arguments_tab.setToolTip(_translate("EmulatorForm", "Contents of arguments if current instruction is a function call"))
        self.function_signature.setToolTip(_translate("EmulatorForm", "Function signature of called function"))
        self.function_signature.setText(_translate("EmulatorForm", "<function signature>"))
        self.num_args_label.setToolTip(_translate("EmulatorForm", "Adjust the number of arguments for the function call"))
        self.num_args_label.setText(_translate("EmulatorForm", "Number of Arguments"))
        self.function_arguments_table.setSortingEnabled(True)
        item = self.function_arguments_table.horizontalHeaderItem(0)
        item.setText(_translate("EmulatorForm", "Ordinal"))
        item = self.function_arguments_table.horizontalHeaderItem(1)
        item.setText(_translate("EmulatorForm", "Location"))
        item = self.function_arguments_table.horizontalHeaderItem(2)
        item.setText(_translate("EmulatorForm", "Data Type"))
        item = self.function_arguments_table.horizontalHeaderItem(3)
        item.setText(_translate("EmulatorForm", "Size"))
        item = self.function_arguments_table.horizontalHeaderItem(4)
        item.setText(_translate("EmulatorForm", "Name"))
        item = self.function_arguments_table.horizontalHeaderItem(5)
        item.setText(_translate("EmulatorForm", "Address"))
        item = self.function_arguments_table.horizontalHeaderItem(6)
        item.setText(_translate("EmulatorForm", "Value"))
        item = self.function_arguments_table.horizontalHeaderItem(7)
        item.setText(_translate("EmulatorForm", "Referenced Data"))
        self.tabs.setTabText(self.tabs.indexOf(self.function_arguments_tab), _translate("EmulatorForm", "Function Arguments"))
        self.call_history_tab.setToolTip(_translate("EmulatorForm", "Function calls observed during emulation"))
        self.call_history_table.setSortingEnabled(True)
        item = self.call_history_table.horizontalHeaderItem(0)
        item.setText(_translate("EmulatorForm", "Address"))
        item = self.call_history_table.horizontalHeaderItem(1)
        item.setText(_translate("EmulatorForm", "Function Call"))
        self.tabs.setTabText(self.tabs.indexOf(self.call_history_tab), _translate("EmulatorForm", "Call History"))
        self.actions_tab.setToolTip(_translate("EmulatorForm", "Interesting actions observed during emulation"))
        self.actions_table.setSortingEnabled(True)
        self.actions_table.headerItem().setText(0, _translate("EmulatorForm", "Address"))
        self.actions_table.headerItem().setText(1, _translate("EmulatorForm", "Action"))
        self.actions_table.headerItem().setText(2, _translate("EmulatorForm", "Attribute"))
        self.actions_table.headerItem().setText(3, _translate("EmulatorForm", "Value"))
        self.tabs.setTabText(self.tabs.indexOf(self.actions_tab), _translate("EmulatorForm", "Actions"))
        self.stdout_tab.setToolTip(_translate("EmulatorForm", "Current contents of the stdout stream"))
        self.stdout_export_button.setText(_translate("EmulatorForm", "Export"))
        self.tabs.setTabText(self.tabs.indexOf(self.stdout_tab), _translate("EmulatorForm", "Stdout"))
        self.status.setText(_translate("EmulatorForm", "<status>"))



from __future__ import annotations
from typing import TYPE_CHECKING

import jpype

RugosaActionListener = jpype.JClass("dc3.rugosa.plugin.RugosaActionListener")

from rugosa.ghidra_plugin.gui.error_msg import catch_errors
from rugosa.ghidra_plugin.gui.table import TableModel


if TYPE_CHECKING:
    from rugosa.ghidra_plugin.components.emulator import EmulatorForm


class CallHistory:

    HEADERS = ["Address", "Function Call"]

    def __init__(self, form: EmulatorForm):
        self._ctx = None
        self.ui = form.ui
        self.table = self.ui.callHistoryTable

    @catch_errors
    def populate(self, ctx):
        """
        Populates the variables table.
        """
        self._ctx = ctx
        rows = []
        for call_address, func_name, args in ctx.call_history:
            args = ", ".join((f"{name}={hex(value)}" if name else hex(value)) for name, value in args)
            func_call = f"{func_name}({args})"
            rows.append([hex(call_address), func_call])
        self.table.setModel(TableModel(self.HEADERS, rows))



from .emulation import Emulator
from .disassembly import *
from .strings import *
# NOTE: Explicit import is necessary to prevent Python from trying to import builtin.
import rugosa.re as re
import rugosa.yara as yara

__version__ = "0.7.2"




from __future__ import annotations

import traceback
from typing import TYPE_CHECKING, Optional

from PyQt5 import QtWidgets, QtCore

import ida_kernwin

import dragodis
import rugosa
from rugosa.ida_plugin.gui.emulator import Ui_EmulatorForm
from rugosa.ida_plugin.components import (
    Operands, Registers, Memory, Variables, FunctionArguments, CallHistory, Actions, Stdout
)

if TYPE_CHECKING:
    from rugosa.emulation.cpu_context import ProcessorContext


class EmulatorForm(Ui_EmulatorForm, ida_kernwin.PluginForm):

    def __init__(self, dis: dragodis.Disassembler):
        super().__init__()
        self._ctx: Optional[ProcessorContext] = None
        self._call_stack = []
        self.emulator = rugosa.Emulator(dis)

    def setupUi(self, form):
        super().setupUi(form)
        # Remove placeholder text.
        self.instruction.setText("")
        self.status.setText("")
        self.function_signature.setText("")

        self.operands = Operands(self)
        self.registers = Registers(self)
        self.memory = Memory(self)
        self.variables = Variables(self)
        self.function_arguments = FunctionArguments(self)
        self.call_history = CallHistory(self)
        self.actions = Actions(self)
        self.stdout = Stdout(self)

        self.instruction.mousePressEvent = self.jump_instruction

        self.run_button.clicked.connect(self.run)
        self.step_over_button.clicked.connect(self.step_over)
        # TODO: Add support for these after refactoring ProcessorContext.
        # self.step_into_button.clicked.connect(self.step_into)
        # self.step_out_button.clicked.connect(self.step_out)

    def OnCreate(self, form):
        """
        Called when the widget is created
        """
        form = self.FormToPyQtWidget(form)
        self.setupUi(form)
        self.parent = form

    def show_exception(self, exception: Exception, title="Error"):
        """
        Shows exception in dialog box.
        """
        dlg = QtWidgets.QMessageBox(self.parent)
        dlg.setIcon(QtWidgets.QMessageBox.Critical)
        dlg.setWindowTitle(title)
        dlg.setText(str(exception))
        details = "\n".join(traceback.format_tb(exception.__traceback__))
        dlg.setDetailedText(details)
        dlg.exec()

    def jump_instruction(self, evt):
        if self._ctx:
            ida_kernwin.jumpto(self._ctx.ip)

    def redirect(self, item: QtWidgets.QTableWidgetItem):
        try:
            address = int(item.text(), 16)
        except ValueError:
            return
        ida_kernwin.jumpto(address)

    def enable_emulator_controls(self):
        """
        Enable emulator buttons based on current context.
        """
        if not self._ctx:
            self.reset()
            return

        self.tabs.setEnabled(True)
        self.step_over_button.setEnabled(True)
        # TODO: Add support for these after some refactoring of ProcessorContext.
        # self.step_into_button.setEnabled(self._ctx.instruction.is_call)
        # self.step_out_button.setEnabled(bool(self._call_stack))

    def update(self):
        """
        Updates UI based on current context.
        """
        ctx = self._ctx
        if not ctx:
            return

        # Enable controls.
        self.enable_emulator_controls()

        insn = ctx.instruction
        self.instruction.setText(f"0x{ctx.ip:08X}: {insn.text}")

        # Fill in tables
        self.operands.populate(ctx)
        self.registers.populate(ctx)
        self.variables.populate(ctx)
        self.function_arguments.populate(ctx)
        self.memory.populate(ctx)
        self.stdout.populate(ctx)
        self.call_history.populate(ctx)
        self.actions.populate(ctx)

        executed_instructions = ctx.executed_instructions
        if executed_instructions:
            self.status.setText(
                f"Emulated {len(executed_instructions)} instructions: "
                f"{hex(ctx.executed_instructions[0])} -> {hex(executed_instructions[-1])}"
            )
        else:
            self.status.setText(f"Emulated 0 instructions.")

        ida_kernwin.jumpto(ctx.ip)

    def run(self):
        """
        Runs emulation using current settings.
        """
        addr = ida_kernwin.get_screen_ea()
        ida_kernwin.show_wait_box(f"HIDECANCEL\nEmulating instructions up to {hex(addr)}")
        try:
            ctx = self.emulator.context_at(
                addr,
                depth=self.trace_depth.value(),
                call_depth=self.call_depth.value(),
                exhaustive=self.exhaustive.isChecked(),
                follow_loops=self.follow_loops.isChecked(),
            )
        except Exception as e:
            ida_kernwin.hide_wait_box()
            self.show_exception(e, "Failed to emulate")
            return
        finally:
            ida_kernwin.hide_wait_box()
        self._ctx = ctx

        self.update()

    def step_over(self):
        ctx = self._ctx
        if not ctx:
            return
        ctx.execute(call_depth=self.call_depth.value())
        self.update()

    # TODO: Need to generate a way to have the context put us in a state where the call instruction is step
    #   (e.g. stack pointer pushed with return address) but not restored (return address popped and set)
    #   - This probably means emulating the 'ret' opcode.
    # def step_into(self):
    #     ctx = self._ctx
    #     if not ctx:
    #         return
    #     try:
    #         func_address = ctx.get_function_signature().address
    #     except dragodis.NotExistError:
    #         self.step_over()  # step over instead if we fail to get the funciton call.
    #         return
    #     except Exception as e:
    #         self.show_exception(e, "Failed to get function call address")
    #         return
    #
    #     # Save call stack info so we can restore it later.
    #     self._call_stack.append((ctx.instruction.next_ip, ctx._call_depth, ctx._sp_start))
    #     ctx.execute(func_address)
    #
    # def step_out(self):
    #     ...

    def reset(self):
        self._ctx = None
        self._call_stack = []
        self.step_over_button.setEnabled(False)
        self.step_into_button.setEnabled(False)
        self.step_out_button.setEnabled(False)
        self.tabs.setEnabled(False)


if __name__ == "__main__":
    form = EmulatorForm(dragodis.IDA())
    form.Show("Rugosa Emulator")



import pytest

import rugosa


def test_find_string_data():
    from rugosa.strings import find_string_data

    # General test of mixture of utf-8 and utf-16
    data = b"h\x00e\x00l\x00l\x00o\x00\x00\x00\x00\x00test\x00\x00\x00joe\x00\x00w\x00o\x00r\x00l\x00d"
    assert list(find_string_data(data)) == [
        (0, b"h\x00e\x00l\x00l\x00o\x00\x00\x00", "utf-16-le"),
        (14, b"test\x00", "utf-8"),
        (21, b"joe\x00", "utf-8"),
        (25, b"\x00w\x00o\x00r\x00l\x00d", "utf-16-be"),
    ]

    # Test arbitrary number of null bytes in between strings.
    data = b"\x00\x00\x00hello\x00\x00\x00\x00\x00\x00w\x00o\x00r\x00l\x00d\x00\x00\x00\x00\x00\x00"
    assert list(find_string_data(data)) == [
        (3, b"hello\x00", "utf-8"),
        (14, b"w\x00o\x00r\x00l\x00d\x00\x00\x00", "utf-16-le"),
    ]

    # Test falling back on single character strings.
    data = b"1\x000\x00\x00\x00hello\x00world\x00"
    assert list(find_string_data(data)) == [
        (0, b"1\x000\x00\x00\x00", "utf-16-le"),
        (6, b"hello\x00", "utf-8"),
        (12, b"world\x00", "utf-8"),
    ]
    data = b"1\x000\x00hello\x00world\x00"
    assert list(find_string_data(data)) == [
        (0, b"1\x00", "utf-8"),
        (2, b"0\x00", "utf-8"),
        (4, b"hello\x00", "utf-8"),
        (10, b"world\x00", "utf-8"),
    ]


def test_force_to_string():
    assert rugosa.force_to_string(b"hello") == "hello"
    assert rugosa.force_to_string("hello".encode("utf-16-be")) == "hello"
    assert rugosa.force_to_string(b"\x4f\xdf\xc6\x4a\xbe\x0a\xff\x76") == "OßÆJ¾\nÿv"


USER_STRINGS = {
    (0x40c000, 'Idmmn!Vnsme '),
    (0x40c010, 'Vgqv"qvpkle"ukvj"ig{"2z20'),
    (0x40c02c, 'Wkf#rvj`h#aqltm#el{#ivnsp#lufq#wkf#obyz#gld-'),
    (0x40c05c, 'Keo$mw$wpvkjc$ej`$ehwk$cmraw$wle`a*'),
    (0x40c080, 'Dfla%gpwkv%mji`v%lk%rjji%fijqm+'),
    (0x40c0a0, 'Egru&ghb&biau&cgen&ngrc&rnc&irnct('),
    # (0x40c0c4, '\cv}3g{v3pargv3qfg3w|}4g3qavrx3g{v3t'),  # TODO: Ghidra fails to find this one.
    (0x40c114, '+()./,-"#*'),
    (0x40c120, '`QFBWFsQL@FPPb'),
    (0x40c130, 'tSUdFS'),
    # (0x40c140, '-",5 , v,tr4v,trv4t,v'),   # TODO: Ghidra fails to find this one.
    (0x40c15c, '@AKJDGBA@KJGDBJKAGDC'),
    (0x40c1f8, 'LMFOGHKNLMGFOHKFGNLKHNMLOKGNKGHFGLHKGLMHKGOFNMLHKGFNLMJNMLIJFGNMLOJIMLNGFJHNM'),
}


API_RESOLVE_STRINGS = {
    (0x40a838, 'KERNEL32.DLL'),
    (0x40a1d4, 'mscoree.dll'),
    (0x40a9d0, 'USER32.DLL'),
    (0x40a1c4, 'CorExitProcess'),
    (0x40a828, 'EncodePointer'),
    (0x40a854, 'DecodePointer'),
    (0x40a884, 'FlsAlloc'),
    (0x40a878, 'FlsGetValue'),
    (0x40a86c, 'FlsSetValue'),
    (0x40a864, 'FlsFree'),
    (0x40a9c4, 'MessageBoxA'),
    (0x40a9b4, 'GetActiveWindow'),
    (0x40a9a0, 'GetLastActivePopup'),
    (0x40a984, 'GetUserObjectInformationA'),
    (0x40a96c, 'GetProcessWindowStation'),
}


def test_find_user_strings(disassembler):
    strings = list(rugosa.find_user_strings(disassembler, unique=True))
    print("\n".join(f"{hex(address)}: '{string}'" for address, string in strings))
    assert set(strings) >= USER_STRINGS
    # While we could have extras based on disassembler, make sure we are somewhat on target.
    assert len(strings) == pytest.approx(len(USER_STRINGS), abs=3)
    assert set(strings).isdisjoint(API_RESOLVE_STRINGS)

    strings = list(rugosa.find_user_strings(disassembler, ignore_api=False, ignore_library=False))
    print("\n".join(f"{hex(address)}: '{string}'" for address, string in strings))
    assert set(strings) >= (USER_STRINGS | API_RESOLVE_STRINGS)


def test_find_api_resolve_strings(disassembler):
    strings = list(rugosa.find_api_resolve_strings(disassembler))
    print("\n".join(f"{hex(address)}: '{string}'" for address, string in strings))
    assert set(strings) == API_RESOLVE_STRINGS


def test_is_library_string(disassembler):
    assert rugosa.is_library_string(disassembler, 0x40A838)
    assert rugosa.is_library_string(disassembler, 0x40A884)
    assert not rugosa.is_library_string(disassembler, 0x40C000)


def test_is_code_string(disassembler):
    assert rugosa.is_code_string(disassembler, 0x40A838)
    assert rugosa.is_code_string(disassembler, 0x40C000)
    assert not rugosa.is_code_string(disassembler, 0x40B2D4)



from __future__ import annotations
from typing import TYPE_CHECKING

from PyQt5 import QtWidgets

if TYPE_CHECKING:
    from rugosa.ida_plugin.components.emulator import EmulatorForm


class Actions:

    def __init__(self, ui: EmulatorForm):
        self._ctx = None
        self.ui = ui
        self.table = ui.actions_table

    def populate(self, ctx):
        """
        Populate the actions table with current context.
        """
        self._ctx = ctx
        actions = ctx.actions
        table = self.table
        table.clear()
        table.setColumnCount(4)
        for action in reversed(actions):
            item = QtWidgets.QTreeWidgetItem([hex(action.ip), action.__class__.__name__, "", ""])
            for name, value in action:
                if name != "ip":
                    item.addChild(QtWidgets.QTreeWidgetItem(["", "", name, str(value)]))
            table.addTopLevelItem(item)



"""
Functions found in shlwapi.dll

Shell Lightweight Utility Functions
"""
import logging
import ntpath

from ...call_hooks import builtin_func


logger = logging.getLogger(__name__)


@builtin_func("PathAppendA")
@builtin_func("PathAppendW")
#typedef(BOOL PathAppendA(LPSTR  pszPath,LPCSTR pszMore);)
def pathappend(cpu_context, func_name, func_args):
    """
    Appends one path to the end of another
    """
    wide = func_name.endswith(u"W")
    path_ptr, more_ptr = func_args

    curr_path = cpu_context.memory.read_string(path_ptr, wide=wide)
    more_path = cpu_context.memory.read_string(more_ptr, wide=wide)

    full_path = ntpath.join(curr_path, more_path)
    cpu_context.memory.write_string(path_ptr, full_path, wide=wide)
    return True


@builtin_func("PathAddBackslashA")
@builtin_func("PathAddBackslashW")
#typedef(LPWSTR PathAddBackslashW(LPWSTR pszPath));)
def pathaddbackslash(cpu_context, func_name, func_args):
    """
    Appends a backslash to the path
    """
    wide = func_name.endswith(u"W")
    path_ptr = func_args[0]

    curr_path = cpu_context.memory.read_string(path_ptr, wide=wide)

    full_path = curr_path + "\\"
    cpu_context.memory.write_string(path_ptr, full_path, wide=wide)
    return True




"""
Interface for instruction management.
"""
import logging

from ..instruction import Instruction
from .operands import x86_64Operand

logger = logging.getLogger(__name__)


class x86_64Instruction(Instruction):

    _operand_class = x86_64Operand

    def _execute(self):
        # Before executing, determine if a rep* instruction and add termination condition.
        term_condition = None
        if rep := self._insn.rep:
            if rep == "rep":
                term_condition = lambda ctx: ctx.registers.ecx == 0
                term_condition.unconditional = True
            elif rep in ("repe", "repz"):
                term_condition = lambda ctx: ctx.registers.ecx == 0 or ctx.registers.zf == 0
                term_condition.unconditional = False
            elif rep in ("repne", "repnz"):
                term_condition = lambda ctx: ctx.registers.ecx == 0 or ctx.registers.zf == 1
                term_condition.unconditional = False
            else:
                logger.warning("Got unexpected rep prefix: %s", self.ip)

        # Execute like normal if not a rep instruciton.
        if not term_condition:
            super()._execute()
            return

        # Skip if user disabled all rep instruction.
        if self._cpu_context.emulator.disabled_rep:
            logger.debug("Ignoring rep instruction at 0x%X: DISABLED.", self.ip)
            return

        # As a safety measure, don't allow rep instructions to surpass
        # our max memory read limit.
        # Only do this check if the terminating condition is unconditional, otherwise
        # this number usually big because it expects zf to be toggled.
        if (
                term_condition.unconditional
                and self._cpu_context.registers.ecx > self._cpu_context.memory.MAX_MEM_READ
        ):
            logger.warning(
                "Emulation attempted to read %s instruction %d times. "
                "Ignoring instruction.", self.mnem, self._cpu_context.registers.ecx
            )
            return

        # Execute instruction ecx number of times or when terminal condition applies.
        logger.debug("Emulating %s instruction %d times.", self.mnem, self._cpu_context.registers.ecx)
        count = 0
        while not term_condition(self._cpu_context):
            super()._execute()
            self._cpu_context.registers.ecx -= 1
            # Stop if we are iterating too much.
            count += 1
            if count > self._cpu_context.memory.MAX_MEM_READ:
                logger.warning("Looped too many times, exiting prematurely.")
                break



from __future__ import annotations
from typing import TYPE_CHECKING

import jpype
from rugosa.ghidra_plugin.gui.error_msg import catch_errors

RugosaActionListener = jpype.JClass("dc3.rugosa.plugin.RugosaActionListener")

if TYPE_CHECKING:
    from java.awt.event import ActionEvent
    from rugosa.ghidra_plugin.components.emulator import EmulatorForm


class Stdout:

    MAX_VALUE_SIZE = 256
    HEADERS = ["Address", "Stack Offset", "Data Type", "Size", "Name", "Value"]

    def __init__(self, form: EmulatorForm):
        self._ctx = None
        self.form = form
        self.ui = form.ui
        self.textdump = self.ui.stdoutTextDump
        self.ui.stdoutExportButton.addActionListener(RugosaActionListener @ self.export)

    @catch_errors
    def export(self, event: ActionEvent):
        """
        Exports current stdout dump to a text file.
        """
        if self._ctx:
            from docking.widgets.filechooser import GhidraFileChooser
            chooser = GhidraFileChooser(self.ui.stdoutTab)
            chooser.setMultiSelectionEnabled(False)
            chooser.setTitle("Export Stdout")
            file = chooser.getSelectedFile()
            if file:
                with open(file.getPath(), "w") as outfile:
                    outfile.write(self._ctx.stdout)

    def populate(self, ctx):
        """
        Populates the variables table.
        """
        self._ctx = ctx
        self.textdump.setText(ctx.stdout)
        self.textdump.scrollToTop()



"""
Contains custom exceptions used to help distinguish emulation errors from
errors caused by bugs.
"""


class EmulationError(Exception):
    def __init__(self, message, ip=None):
        super().__init__(message)
        self.ip = ip

    def __str__(self):
        message = super().__str__()
        if self.ip is not None:
            return f"0x{self.ip:X} :: " + message
        else:
            return message
